from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    NumericCritic,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_spotify.tools.player import (
    adjust_playback_position,
    get_currently_playing,
    get_playback_state,
    pause_playback,
    play_artist_by_name,
    play_track_by_name,
    resume_playback,
    skip_to_next_track,
    skip_to_previous_track,
    start_tracks_playback_by_id,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_tool(adjust_playback_position, "Spotify")
catalog.add_tool(skip_to_next_track, "Spotify")
catalog.add_tool(skip_to_previous_track, "Spotify")
catalog.add_tool(pause_playback, "Spotify")
catalog.add_tool(resume_playback, "Spotify")
catalog.add_tool(start_tracks_playback_by_id, "Spotify")
catalog.add_tool(get_playback_state, "Spotify")
catalog.add_tool(get_currently_playing, "Spotify")
catalog.add_tool(play_artist_by_name, "Spotify")
catalog.add_tool(play_track_by_name, "Spotify")


@tool_eval()
def spotify_player_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Spotify "player" tools."""
    suite = EvalSuite(
        name="Spotify Tools Evaluation",
        system_message="You are an AI assistant that can manage Spotify using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Adjust playback position",
        user_message="can you skip to the 10th second of the song",
        expected_tool_calls=[
            ExpectedToolCall(
                func=adjust_playback_position,
                args={"absolute_position_ms": 10000},
            )
        ],
        critics=[
            NumericCritic(
                critic_field="absolute_position_ms", weight=1.0, value_range=(9000, 11000)
            ),
        ],
    )

    suite.add_case(
        name="Adjust playback position relative to current position",
        user_message="go back 10 seconds",
        expected_tool_calls=[
            ExpectedToolCall(
                func=adjust_playback_position,
                args={"relative_position_ms": -10000},
            )
        ],
        critics=[
            NumericCritic(
                critic_field="relative_position_ms",
                weight=1.0,
                value_range=(-11000, -9000),
            ),
        ],
    )

    suite.add_case(
        name="Skip to previous track",
        user_message="oops i didn't mean to skip that song, go back",
        expected_tool_calls=[ExpectedToolCall(func=skip_to_previous_track, args={})],
    )

    suite.add_case(
        name="Skip to next track",
        user_message="skip this song and also the next one",
        expected_tool_calls=[
            ExpectedToolCall(func=skip_to_next_track, args={}),
            ExpectedToolCall(func=skip_to_next_track, args={}),
        ],
    )

    suite.add_case(
        name="Pause playback",
        user_message="wait im getting a text, stop playing it please",
        expected_tool_calls=[ExpectedToolCall(func=pause_playback, args={})],
    )

    suite.add_case(
        name="Resume playback",
        user_message="ok i'm back, you can press play again",
        expected_tool_calls=[ExpectedToolCall(func=resume_playback, args={})],
    )

    suite.add_case(
        name="Start playback of a list of tracks",
        user_message="Play these two 03gaqN3aWm9TQxuHay0G8R, 03gaqN3aWm9TQxuHay0G8R. But start at the 10th second of the first track",
        expected_tool_calls=[
            ExpectedToolCall(
                func=start_tracks_playback_by_id,
                args={
                    "track_ids": ["03gaqN3aWm9TQxuHay0G8R", "03gaqN3aWm9TQxuHay0G8R"],
                    "position_ms": 10000,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="track_ids", weight=0.5),
            NumericCritic(critic_field="position_ms", weight=0.5, value_range=(9000, 11000)),
        ],
    )

    suite.add_case(
        name="Get playback state",
        user_message="what's the name of this song and who plays it?",
        expected_tool_calls=[ExpectedToolCall(func=get_currently_playing, args={})],
    )

    suite.add_case(
        name="Get playback state",
        user_message="what device is playing music rn?",
        expected_tool_calls=[ExpectedToolCall(func=get_playback_state, args={})],
    )

    suite.add_case(
        name="Play artist by name",
        user_message="play pearl jam",
        expected_tool_calls=[
            ExpectedToolCall(
                func=play_artist_by_name,
                args={"name": "Pearl Jam"},
            )
        ],
        critics=[
            SimilarityCritic(critic_field="name", weight=1.0),
        ],
    )

    suite.add_case(
        name="Play track by name",
        user_message="it would be really great if I could listen to strobe by deadmau5 right now.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=play_track_by_name,
                args={"track_name": "strobe", "artist_name": "deadmau5"},
            )
        ],
        critics=[
            SimilarityCritic(critic_field="track_name", weight=0.5),
            SimilarityCritic(critic_field="artist_name", weight=0.5),
        ],
    )

    return suite



import os
from enum import Enum

ASANA_BASE_URL = "https://app.asana.com/api"
ASANA_API_VERSION = "1.0"

try:
    ASANA_MAX_CONCURRENT_REQUESTS = int(os.getenv("ASANA_MAX_CONCURRENT_REQUESTS", 3))
except ValueError:
    ASANA_MAX_CONCURRENT_REQUESTS = 3

try:
    ASANA_MAX_TIMEOUT_SECONDS = int(os.getenv("ASANA_MAX_TIMEOUT_SECONDS", 20))
except ValueError:
    ASANA_MAX_TIMEOUT_SECONDS = 20

MAX_PROJECTS_TO_SCAN_BY_NAME = 1000
MAX_TAGS_TO_SCAN_BY_NAME = 1000

PROJECT_OPT_FIELDS = [
    "gid",
    "resource_type",
    "name",
    "workspace",
    "color",
    "created_at",
    "current_status_update",
    "due_on",
    "members",
    "notes",
    "completed",
    "completed_at",
    "completed_by",
    "owner",
    "team",
    "workspace",
    "permalink_url",
]

TASK_OPT_FIELDS = [
    "gid",
    "name",
    "notes",
    "completed",
    "completed_at",
    "completed_by",
    "created_at",
    "created_by",
    "due_on",
    "start_on",
    "owner",
    "team",
    "workspace",
    "permalink_url",
    "approval_status",
    "assignee",
    "assignee_status",
    "dependencies",
    "dependents",
    "memberships",
    "num_subtasks",
    "resource_type",
    "custom_type",
    "custom_type_status_option",
    "parent",
    "tags",
    "workspace",
]


TAG_OPT_FIELDS = [
    "gid",
    "name",
    "workspace",
]

TEAM_OPT_FIELDS = [
    "gid",
    "name",
    "description",
    "organization",
    "permalink_url",
]

USER_OPT_FIELDS = [
    "gid",
    "resource_type",
    "name",
    "email",
    "photo",
    "workspaces",
]

WORKSPACE_OPT_FIELDS = [
    "gid",
    "resource_type",
    "name",
    "email_domains",
    "is_organization",
]


class TaskSortBy(Enum):
    DUE_DATE = "due_date"
    CREATED_AT = "created_at"
    COMPLETED_AT = "completed_at"
    MODIFIED_AT = "modified_at"
    LIKES = "likes"


class SortOrder(Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class TagColor(Enum):
    DARK_GREEN = "dark-green"
    DARK_RED = "dark-red"
    DARK_BLUE = "dark-blue"
    DARK_PURPLE = "dark-purple"
    DARK_PINK = "dark-pink"
    DARK_ORANGE = "dark-orange"
    DARK_TEAL = "dark-teal"
    DARK_BROWN = "dark-brown"
    DARK_WARM_GRAY = "dark-warm-gray"
    LIGHT_GREEN = "light-green"
    LIGHT_RED = "light-red"
    LIGHT_BLUE = "light-blue"
    LIGHT_PURPLE = "light-purple"
    LIGHT_PINK = "light-pink"
    LIGHT_ORANGE = "light-orange"
    LIGHT_TEAL = "light-teal"
    LIGHT_BROWN = "light-brown"
    LIGHT_WARM_GRAY = "light-warm-gray"


class ReturnType(Enum):
    FULL_ITEMS_DATA = "full_items_data"
    ITEMS_COUNT = "items_count"



from datetime import datetime, timezone
from typing import Any

from arcade_tdk.errors import ToolExecutionError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.utils import (
    async_paginate,
    convert_datetime_to_unix_timestamp,
    convert_relative_datetime_to_unix_timestamp,
    enrich_message_datetime,
)


async def retrieve_messages_in_conversation(
    conversation_id: str,
    auth_token: str | None = None,
    oldest_relative: str | None = None,
    latest_relative: str | None = None,
    oldest_datetime: str | None = None,
    latest_datetime: str | None = None,
    limit: int | None = None,
    next_cursor: str | None = None,
) -> dict:
    error_message = None
    if oldest_datetime and oldest_relative:
        error_message = "Cannot specify both 'oldest_datetime' and 'oldest_relative'."

    if latest_datetime and latest_relative:
        error_message = "Cannot specify both 'latest_datetime' and 'latest_relative'."

    if error_message:
        raise ToolExecutionError(error_message, developer_message=error_message)

    current_unix_timestamp = int(datetime.now(timezone.utc).timestamp())

    if latest_relative:
        latest_timestamp = convert_relative_datetime_to_unix_timestamp(
            latest_relative, current_unix_timestamp
        )
    elif latest_datetime:
        latest_timestamp = convert_datetime_to_unix_timestamp(latest_datetime)
    else:
        latest_timestamp = None

    if oldest_relative:
        oldest_timestamp = convert_relative_datetime_to_unix_timestamp(
            oldest_relative, current_unix_timestamp
        )
    elif oldest_datetime:
        oldest_timestamp = convert_datetime_to_unix_timestamp(oldest_datetime)
    else:
        oldest_timestamp = None

    datetime_args: dict[str, Any] = {}
    if oldest_timestamp:
        datetime_args["oldest"] = oldest_timestamp
    if latest_timestamp:
        datetime_args["latest"] = latest_timestamp

    slackClient = AsyncWebClient(token=auth_token)

    response, next_cursor = await async_paginate(
        slackClient.conversations_history,
        "messages",
        limit=limit,
        next_cursor=next_cursor,
        channel=conversation_id,
        include_all_metadata=True,
        inclusive=True,  # Include messages at the start and end of the time range
        **datetime_args,
    )

    messages = [enrich_message_datetime(message) for message in response]

    return {"messages": messages, "next_cursor": next_cursor}



from typing import Optional


class DropboxApiError(Exception):
    def __init__(
        self,
        status_code: int,
        error_summary: str,
        user_message: Optional[str],
    ):
        if "path/not_found" in error_summary:
            self.message = "The specified path was not found by Dropbox"
        elif "unsupported_file" in error_summary:
            self.message = "The specified file is not supported for the requested operation"
        else:
            self.message = user_message or error_summary

        self.status_code = status_code



from unittest.mock import MagicMock

import pytest

from arcade_spotify.tools.models import PlaybackState
from arcade_spotify.tools.utils import convert_to_playback_state, send_spotify_request


@pytest.mark.asyncio
async def test_send_spotify_request(tool_context, mock_httpx_client):
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_httpx_client.request.return_value = mock_response

    response = await send_spotify_request(
        tool_context,
        "GET",
        "https://api.spotify.com/v1/me/player",
        params={"param": "value"},
        json_data={"data": "value"},
    )
    assert response == mock_response
    mock_httpx_client.request.assert_called_once_with(
        "GET",
        "https://api.spotify.com/v1/me/player",
        headers={"Authorization": "Bearer test_token"},
        params={"param": "value"},
        json={"data": "value"},
    )


def test_convert_to_playback_state():
    player_get_playback_state_response = {
        "timestamp": 1734651060828,
        "context": {
            "external_urls": {
                "spotify": "https://open.spotify.com/playlist/37i9dQZF1EYkqdzj48dyYq"
            },
            "href": "https://api.spotify.com/v1/playlists/37i9dQZF1EYkqdzj48dyYq",
            "type": "playlist",
            "uri": "spotify:playlist:37i9dQZF1EYkqdzj48dyYq",
        },
        "progress_ms": 261652,
        "item": {
            "album": {
                "album_type": "album",
                "artists": [
                    {
                        "external_urls": {
                            "spotify": "https://open.spotify.com/artist/3GBPw9NK25X1Wt2OUvOwY3"
                        },
                        "href": "https://api.spotify.com/v1/artists/3GBPw9NK25X1Wt2OUvOwY3",
                        "id": "3GBPw9NK25X1Wt2OUvOwY3",
                        "name": "Jack Johnson",
                        "type": "artist",
                        "uri": "spotify:artist:3GBPw9NK25X1Wt2OUvOwY3",
                    }
                ],
                "available_markets": [
                    "AR",
                    "XK",
                ],
                "external_urls": {
                    "spotify": "https://open.spotify.com/album/23BBbqDGMhloT6f2YBecSr"
                },
                "href": "https://api.spotify.com/v1/albums/23BBbqDGMhloT6f2YBecSr",
                "id": "23BBbqDGMhloT6f2YBecSr",
                "images": [
                    {
                        "height": 640,
                        "url": "https://i.scdn.co/image/ab67616d0000b2732bd026ab797a3de9605d9cb3",
                        "width": 640,
                    },
                    {
                        "height": 300,
                        "url": "https://i.scdn.co/image/ab67616d00001e022bd026ab797a3de9605d9cb3",
                        "width": 300,
                    },
                    {
                        "height": 64,
                        "url": "https://i.scdn.co/image/ab67616d000048512bd026ab797a3de9605d9cb3",
                        "width": 64,
                    },
                ],
                "name": "Brushfire Fairytales [Remastered (Bonus Version)]",
                "release_date": "2011-04-12",
                "release_date_precision": "day",
                "total_tracks": 15,
                "type": "album",
                "uri": "spotify:album:23BBbqDGMhloT6f2YBecSr",
            },
            "artists": [
                {
                    "external_urls": {
                        "spotify": "https://open.spotify.com/artist/3GBPw9NK25X1Wt2OUvOwY3"
                    },
                    "href": "https://api.spotify.com/v1/artists/3GBPw9NK25X1Wt2OUvOwY3",
                    "id": "3GBPw9NK25X1Wt2OUvOwY3",
                    "name": "Jack Johnson",
                    "type": "artist",
                    "uri": "spotify:artist:3GBPw9NK25X1Wt2OUvOwY3",
                }
            ],
            "available_markets": [
                "AR",
                "XK",
            ],
            "disc_number": 1,
            "duration_ms": 281749,
            "explicit": False,
            "external_ids": {"isrc": "USER81100105"},
            "external_urls": {"spotify": "https://open.spotify.com/track/54S3uCvfZauNw8lVCHZYYo"},
            "href": "https://api.spotify.com/v1/tracks/54S3uCvfZauNw8lVCHZYYo",
            "id": "54S3uCvfZauNw8lVCHZYYo",
            "is_local": False,
            "name": "Flake",
            "popularity": 59,
            "preview_url": "https://p.scdn.co/mp3-preview/7094898f5aa76222b06349e4ec26489ca80b5e4f?cid=26913f34d26f4c16a15d5a93e309a1dc",
            "track_number": 5,
            "type": "track",
            "uri": "spotify:track:54S3uCvfZauNw8lVCHZYYo",
        },
        "currently_playing_type": "track",
        "actions": {
            "disallows": {
                "resuming": True,
                "toggling_repeat_context": True,
                "toggling_repeat_track": True,
                "toggling_shuffle": True,
            }
        },
        "is_playing": True,
    }

    expected_playback_state = PlaybackState(
        device_name=None,
        device_id=None,
        currently_playing_type="track",
        is_playing=True,
        progress_ms=261652,
        message=None,
        album_name="Brushfire Fairytales [Remastered (Bonus Version)]",
        album_id="23BBbqDGMhloT6f2YBecSr",
        album_artists=["Jack Johnson"],
        album_spotify_url="https://open.spotify.com/album/23BBbqDGMhloT6f2YBecSr",
        track_name="Flake",
        track_id="54S3uCvfZauNw8lVCHZYYo",
        track_spotify_url="https://open.spotify.com/track/54S3uCvfZauNw8lVCHZYYo",
        track_artists=["Jack Johnson"],
        track_artists_ids=["3GBPw9NK25X1Wt2OUvOwY3"],
    )

    playback_state = convert_to_playback_state(player_get_playback_state_response)

    assert playback_state == expected_playback_state



import re
from typing import Any

# TODO: This is a partial implementation.
# TODO: Does not support children blocks. Instead, the markdown content is flattened.
# TODO: Does not support equation blocks.
# TODO: Does not support colored text styling.
# TODO: Does not support underline text styling.
# TODO: Does not support multiple text styles for the same block.


def convert_markdown_to_blocks(content: str) -> list[dict[str, Any]]:  # noqa: C901
    """Convert markdown content to Notion blocks."""
    blocks: list[dict[str, Any]] = []
    code_block: list[str] = []
    in_code: bool = False
    language: str = "plain text"
    numbered_list_index: int = 0

    for line in content.splitlines():
        line = line.strip()

        if line.startswith("```"):
            if in_code:
                blocks.append({
                    "type": "code",
                    "code": {
                        "rich_text": [
                            {
                                "type": "text",
                                "text": {"content": "\n".join(code_block)},
                            }
                        ],
                        "language": language,
                    },
                })
                code_block = []
                in_code = False
            else:
                in_code = True
                language = line[3:].strip() or "plain text"
            continue

        if in_code:
            code_block.append(line)
            continue

        if not line:
            numbered_list_index = 0
            continue

        if line.startswith("### "):
            block_type, text = "heading_3", line[4:]
        elif line.startswith("## "):
            block_type, text = "heading_2", line[3:]
        elif line.startswith("# "):
            block_type, text = "heading_1", line[2:]
        elif numbered_match := re.match(r"(\d+)\.\s+(.+)", line):
            block_type, text = "numbered_list_item", numbered_match.group(2)
            numbered_list_index += 1
        elif line.startswith("- "):
            block_type, text = "bulleted_list_item", line[2:]
        elif line.startswith("> "):
            block_type, text = "quote", line[2:]
        elif line == "---":
            blocks.append({"type": "divider", "divider": {}})
            continue
        else:
            block_type, text = "paragraph", line

        blocks.append({"type": block_type, block_type: {"rich_text": format_text(text)}})

    return blocks


def format_text(text: str) -> list[dict[str, Any]]:
    """Convert text with markdown formatting to Notion rich text."""
    patterns = [
        (r"\[([^\]]+)\]\(([^\)]+)\)", "link"),  # [text](url)
        (r"\*\*(.*?)\*\*", "bold"),  # **bold**
        (r"__(.*?)__", "bold"),  # __bold__
        (r"\*(.*?)\*", "italic"),  # *italic*
        (r"_(.*?)_", "italic"),  # _italic_
        (r"~~(.*?)~~", "strikethrough"),  # ~~strikethrough~~
        (r"`(.*?)`", "code"),  # `code`
    ]

    rich_text = []
    last_index = 0
    combined_pattern = "|".join(f"({pattern})" for pattern, _ in patterns)

    for match in re.finditer(combined_pattern, text):
        start, end = match.span()
        if start > last_index:
            rich_text.append({
                "type": "text",
                "text": {"content": text[last_index:start]},
                "annotations": {
                    "bold": False,
                    "italic": False,
                    "strikethrough": False,
                    "underline": False,
                    "code": False,
                    "color": "default",
                },
            })

        matched_text = match.group(0)
        for pattern, format_type in patterns:
            if m := re.match(pattern, matched_text):
                if format_type == "link":
                    rich_text.append({
                        "type": "text",
                        "text": {
                            "content": m.group(1),
                            "link": {"url": m.group(2)},
                        },
                        "annotations": {
                            "bold": False,
                            "italic": False,
                            "strikethrough": False,
                            "underline": False,
                            "code": False,
                            "color": "default",
                        },
                    })
                else:
                    rich_text.append({
                        "type": "text",
                        "text": {"content": m.group(1)},
                        "annotations": {
                            "bold": format_type == "bold",
                            "italic": format_type == "italic",
                            "strikethrough": format_type == "strikethrough",
                            "underline": False,
                            "code": format_type == "code",
                            "color": "default",
                        },
                    })
                break

        last_index = end

    if last_index < len(text):
        rich_text.append({
            "type": "text",
            "text": {"content": text[last_index:]},
            "annotations": {
                "bold": False,
                "italic": False,
                "strikethrough": False,
                "underline": False,
                "code": False,
                "color": "default",
            },
        })

    return rich_text



import os

YOUTUBE_MAX_DESCRIPTION_LENGTH = 500
DEFAULT_YOUTUBE_SEARCH_LANGUAGE = os.getenv("ARCADE_YOUTUBE_SEARCH_LANGUAGE")
DEFAULT_YOUTUBE_SEARCH_COUNTRY = os.getenv("ARCADE_YOUTUBE_SEARCH_COUNTRY")
DEFAULT_GOOGLE_LANGUAGE = os.getenv("ARCADE_GOOGLE_LANGUAGE", "en")
DEFAULT_GOOGLE_COUNTRY = os.getenv("ARCADE_GOOGLE_COUNTRY")



from collections.abc import Callable

import pytest
from arcade_tdk import ToolContext

from arcade_jira.exceptions import MultipleItemsFoundError, NotFoundError
from arcade_jira.utils import (
    clean_user_dict,
    find_multiple_unique_users,
    find_unique_user,
)


@pytest.mark.asyncio
async def test_find_unique_user_by_id_success(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    fake_cloud_name: str,
):
    sample_user = build_user_dict()
    user_response = mock_httpx_response(200, sample_user)
    mock_httpx_client.get.return_value = user_response

    response = await find_unique_user(mock_context, sample_user["accountId"])
    assert response == clean_user_dict(sample_user, fake_cloud_name)


@pytest.mark.asyncio
async def test_find_unique_user_by_name_with_a_single_match(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    fake_cloud_name: str,
):
    sample_user = build_user_dict()
    get_user_by_id_response = mock_httpx_response(404, {})
    get_users_without_id_response = mock_httpx_response(200, [sample_user])
    mock_httpx_client.get.side_effect = [get_user_by_id_response, get_users_without_id_response]

    response = await find_unique_user(mock_context, sample_user["displayName"].lower())
    assert response == clean_user_dict(sample_user, fake_cloud_name)


@pytest.mark.asyncio
async def test_find_unique_user_by_name_with_multiple_matches(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    generate_random_str: Callable,
):
    user_name = generate_random_str()
    sample_users = [
        build_user_dict(display_name=user_name),
        build_user_dict(display_name=user_name),
    ]
    get_user_by_id_response = mock_httpx_response(404, {})
    get_users_without_id_response = mock_httpx_response(200, sample_users)
    mock_httpx_client.get.side_effect = [get_user_by_id_response, get_users_without_id_response]

    with pytest.raises(MultipleItemsFoundError) as exc:
        await find_unique_user(mock_context, sample_users[0]["displayName"].lower())

    assert sample_users[0]["accountId"] in exc.value.message
    assert sample_users[1]["accountId"] in exc.value.message


@pytest.mark.asyncio
async def test_find_unique_user_by_name_without_matches(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    generate_random_str: Callable,
):
    get_user_by_id_response = mock_httpx_response(404, {})
    get_users_without_id_response = mock_httpx_response(200, [])
    mock_httpx_client.get.side_effect = [get_user_by_id_response, get_users_without_id_response]

    with pytest.raises(NotFoundError):
        await find_unique_user(mock_context, generate_random_str())


@pytest.mark.asyncio
async def test_find_multiple_users_when_all_names_match_one_result(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    fake_cloud_name: str,
):
    user1 = build_user_dict()
    user2 = build_user_dict()

    mock_httpx_client.get.side_effect = [
        mock_httpx_response(200, [user1]),
        mock_httpx_response(200, [user2]),
    ]

    response = await find_multiple_unique_users(
        mock_context, [user1["displayName"], user2["displayName"]]
    )

    assert response == [
        clean_user_dict(user1, fake_cloud_name),
        clean_user_dict(user2, fake_cloud_name),
    ]


@pytest.mark.asyncio
async def test_find_multiple_users_when_a_name_match_multiple_results(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
):
    user1 = build_user_dict()
    user2 = build_user_dict()
    user3 = build_user_dict()

    mock_httpx_client.get.side_effect = [
        mock_httpx_response(200, [user1]),
        mock_httpx_response(200, [user2, user3]),
    ]

    with pytest.raises(MultipleItemsFoundError) as exc:
        await find_multiple_unique_users(mock_context, [user1["displayName"], user2["displayName"]])

    assert user2["accountId"] in exc.value.message
    assert user3["accountId"] in exc.value.message


@pytest.mark.asyncio
async def test_find_multiple_users_when_user_is_not_found_by_name_but_found_by_id(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    fake_cloud_name: str,
):
    user1 = build_user_dict()
    user2 = build_user_dict()

    mock_httpx_client.get.side_effect = [
        mock_httpx_response(200, [user1]),
        mock_httpx_response(200, []),
        mock_httpx_response(200, user2),
    ]

    response = await find_multiple_unique_users(
        mock_context, [user1["displayName"], user2["accountId"]]
    )

    assert response == [
        clean_user_dict(user1, fake_cloud_name),
        clean_user_dict(user2, fake_cloud_name),
    ]


@pytest.mark.asyncio
async def test_find_multiple_users_when_various_users_are_not_found_by_name_but_found_by_id(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_user_dict: Callable,
    fake_cloud_name: str,
):
    user1 = build_user_dict()
    user2 = build_user_dict()
    user3 = build_user_dict()

    mock_httpx_client.get.side_effect = [
        mock_httpx_response(200, [user1]),
        mock_httpx_response(200, []),
        mock_httpx_response(200, []),
        mock_httpx_response(200, user2),
        mock_httpx_response(200, user3),
    ]

    response = await find_multiple_unique_users(
        mock_context, [user1["displayName"], user2["accountId"], user3["accountId"]]
    )

    assert response == [
        clean_user_dict(user1, fake_cloud_name),
        clean_user_dict(user2, fake_cloud_name),
        clean_user_dict(user3, fake_cloud_name),
    ]



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Google
from googleapiclient.errors import HttpError

from arcade_google.doc_to_html import convert_document_to_html
from arcade_google.doc_to_markdown import convert_document_to_markdown
from arcade_google.models import DocumentFormat, OrderBy
from arcade_google.tools import get_document_by_id
from arcade_google.utils import (
    build_drive_service,
    build_file_tree,
    build_file_tree_request_params,
    build_files_list_params,
)


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    )
)
async def get_file_tree_structure(
    context: ToolContext,
    include_shared_drives: Annotated[
        bool, "Whether to include shared drives in the file tree structure. Defaults to False."
    ] = False,
    restrict_to_shared_drive_id: Annotated[
        str | None,
        "If provided, only include files from this shared drive in the file tree structure. "
        "Defaults to None, which will include files and folders from all drives.",
    ] = None,
    include_organization_domain_documents: Annotated[
        bool,
        "Whether to include documents from the organization's domain. This is applicable to admin "
        "users who have permissions to view organization-wide documents in a Google Workspace "
        "account. Defaults to False.",
    ] = False,
    order_by: Annotated[
        list[OrderBy] | None,
        "Sort order. Defaults to listing the most recently modified documents first",
    ] = None,
    limit: Annotated[
        int | None,
        "The number of files and folders to list. Defaults to None, "
        "which will list all files and folders.",
    ] = None,
) -> Annotated[
    dict,
    "A dictionary containing the file/folder tree structure in the user's Google Drive",
]:
    """
    Get the file/folder tree structure of the user's Google Drive.
    """
    service = build_drive_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    keep_paginating = True
    page_token = None
    files = {}
    file_tree: dict[str, list[dict]] = {"My Drive": []}

    params = build_file_tree_request_params(
        order_by,
        page_token,
        limit,
        include_shared_drives,
        restrict_to_shared_drive_id,
        include_organization_domain_documents,
    )

    while keep_paginating:
        # Get a list of files
        results = service.files().list(**params).execute()

        # Update page token
        page_token = results.get("nextPageToken")
        params["pageToken"] = page_token
        keep_paginating = page_token is not None

        for file in results.get("files", []):
            files[file["id"]] = file

    if not files:
        return {"drives": []}

    file_tree = build_file_tree(files)

    drives = []

    for drive_id, files in file_tree.items():  # type: ignore[assignment]
        if drive_id == "My Drive":
            drive = {"name": "My Drive", "children": files}
        else:
            try:
                drive_details = service.drives().get(driveId=drive_id).execute()
                drive_name = drive_details.get("name", "Shared Drive (name unavailable)")
            except HttpError as e:
                drive_name = (
                    f"Shared Drive (name unavailable: 'HttpError {e.status_code}: {e.reason}')"
                )

            drive = {"name": drive_name, "id": drive_id, "children": files}

        drives.append(drive)

    return {"drives": drives}


# Implements: https://googleapis.github.io/google-api-python-client/docs/dyn/drive_v3.files.html#list
# Example `arcade chat` query: `list my 5 most recently modified documents`
# TODO: Support query with natural language. Currently, the tool expects a fully formed query
#       string as input with the syntax defined here: https://developers.google.com/drive/api/guides/search-files
@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    )
)
async def search_documents(
    context: ToolContext,
    document_contains: Annotated[
        list[str] | None,
        "Keywords or phrases that must be in the document title or body. Provide a list of "
        "keywords or phrases if needed.",
    ] = None,
    document_not_contains: Annotated[
        list[str] | None,
        "Keywords or phrases that must NOT be in the document title or body. Provide a list of "
        "keywords or phrases if needed.",
    ] = None,
    search_only_in_shared_drive_id: Annotated[
        str | None,
        "The ID of the shared drive to restrict the search to. If provided, the search will only "
        "return documents from this drive. Defaults to None, which searches across all drives.",
    ] = None,
    include_shared_drives: Annotated[
        bool,
        "Whether to include documents from shared drives. Defaults to False (searches only in "
        "the user's 'My Drive').",
    ] = False,
    include_organization_domain_documents: Annotated[
        bool,
        "Whether to include documents from the organization's domain. This is applicable to admin "
        "users who have permissions to view organization-wide documents in a Google Workspace "
        "account. Defaults to False.",
    ] = False,
    order_by: Annotated[
        list[OrderBy] | None,
        "Sort order. Defaults to listing the most recently modified documents first",
    ] = None,
    limit: Annotated[int, "The number of documents to list"] = 50,
    pagination_token: Annotated[
        str | None, "The pagination token to continue a previous request"
    ] = None,
) -> Annotated[
    dict,
    "A dictionary containing 'documents_count' (number of documents returned) and 'documents' "
    "(a list of document details including 'kind', 'mimeType', 'id', and 'name' for each document)",
]:
    """
    Searches for documents in the user's Google Drive. Excludes documents that are in the trash.
    """
    if order_by is None:
        order_by = [OrderBy.MODIFIED_TIME_DESC]
    elif isinstance(order_by, OrderBy):
        order_by = [order_by]

    page_size = min(10, limit)
    files: list[dict[str, Any]] = []

    service = build_drive_service(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    params = build_files_list_params(
        mime_type="application/vnd.google-apps.document",
        document_contains=document_contains,
        document_not_contains=document_not_contains,
        page_size=page_size,
        order_by=order_by,
        pagination_token=pagination_token,
        include_shared_drives=include_shared_drives,
        search_only_in_shared_drive_id=search_only_in_shared_drive_id,
        include_organization_domain_documents=include_organization_domain_documents,
    )

    while len(files) < limit:
        if pagination_token:
            params["pageToken"] = pagination_token
        else:
            params.pop("pageToken", None)

        results = service.files().list(**params).execute()
        batch = results.get("files", [])
        files.extend(batch[: limit - len(files)])

        pagination_token = results.get("nextPageToken")
        if not pagination_token or len(batch) < page_size:
            break

    return {"documents_count": len(files), "documents": files}


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    )
)
async def search_and_retrieve_documents(
    context: ToolContext,
    return_format: Annotated[
        DocumentFormat,
        "The format of the document to return. Defaults to Markdown.",
    ] = DocumentFormat.MARKDOWN,
    document_contains: Annotated[
        list[str] | None,
        "Keywords or phrases that must be in the document title or body. Provide a list of "
        "keywords or phrases if needed.",
    ] = None,
    document_not_contains: Annotated[
        list[str] | None,
        "Keywords or phrases that must NOT be in the document title or body. Provide a list of "
        "keywords or phrases if needed.",
    ] = None,
    search_only_in_shared_drive_id: Annotated[
        str | None,
        "The ID of the shared drive to restrict the search to. If provided, the search will only "
        "return documents from this drive. Defaults to None, which searches across all drives.",
    ] = None,
    include_shared_drives: Annotated[
        bool,
        "Whether to include documents from shared drives. Defaults to False (searches only in "
        "the user's 'My Drive').",
    ] = False,
    include_organization_domain_documents: Annotated[
        bool,
        "Whether to include documents from the organization's domain. This is applicable to admin "
        "users who have permissions to view organization-wide documents in a Google Workspace "
        "account. Defaults to False.",
    ] = False,
    order_by: Annotated[
        list[OrderBy] | None,
        "Sort order. Defaults to listing the most recently modified documents first",
    ] = None,
    limit: Annotated[int, "The number of documents to list"] = 50,
    pagination_token: Annotated[
        str | None, "The pagination token to continue a previous request"
    ] = None,
) -> Annotated[
    dict,
    "A dictionary containing 'documents_count' (number of documents returned) and 'documents' "
    "(a list of documents with their content).",
]:
    """
    Searches for documents in the user's Google Drive and returns a list of documents (with text
    content) matching the search criteria. Excludes documents that are in the trash.

    Note: use this tool only when the user prompt requires the documents' content. If the user only
    needs a list of documents, use the `search_documents` tool instead.
    """
    response = await search_documents(
        context=context,
        document_contains=document_contains,
        document_not_contains=document_not_contains,
        search_only_in_shared_drive_id=search_only_in_shared_drive_id,
        include_shared_drives=include_shared_drives,
        include_organization_domain_documents=include_organization_domain_documents,
        order_by=order_by,
        limit=limit,
        pagination_token=pagination_token,
    )

    documents = []

    for item in response["documents"]:
        document = await get_document_by_id(context, document_id=item["id"])

        if return_format == DocumentFormat.MARKDOWN:
            document = convert_document_to_markdown(document)
        elif return_format == DocumentFormat.HTML:
            document = convert_document_to_html(document)

        documents.append(document)

    return {"documents_count": len(documents), "documents": documents}



import json

from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_asana
from arcade_asana.tools import create_tag, list_tags

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_asana)


@tool_eval()
def list_tags_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="list tags eval suite",
        system_message=(
            "You are an AI assistant with access to Asana tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List tags",
        user_message="List the tags in Asana.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_tags,
                args={
                    "limit": 100,
                    "offset": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
    )

    suite.add_case(
        name="List tags with limit",
        user_message="List 10 tags in Asana.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_tags,
                args={
                    "limit": 10,
                    "offset": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.75),
            BinaryCritic(critic_field="offset", weight=0.25),
        ],
    )

    suite.add_case(
        name="List tags with pagination",
        user_message="Show me the next 2 tags.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_tags,
                args={
                    "limit": 2,
                    "offset": "abc123",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
        additional_messages=[
            {"role": "user", "content": "Show me 2 tags in Asana."},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Asana_ListTags",
                            "arguments": '{"limit":2}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "count": 2,
                    "next_page": {
                        "has_more_results": True,
                        "next_page_token": "abc123",
                    },
                    "workspaces": [
                        {
                            "id": "1234567890",
                            "name": "Tag Hello",
                        },
                        {
                            "id": "1234567891",
                            "name": "Tag World",
                        },
                    ],
                }),
                "tool_call_id": "call_1",
                "name": "Asana_ListTags",
            },
            {
                "role": "assistant",
                "content": "Here are two tags in Asana:\n\n1. Tag Hello\n2. Tag World",
            },
        ],
    )

    suite.add_case(
        name="List tags with pagination changing the limit",
        user_message="Show me the next 5 tags.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_tags,
                args={
                    "limit": 5,
                    "offset": "abc123",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="offset", weight=0.5),
        ],
        additional_messages=[
            {"role": "user", "content": "Show me 5 tags in Asana."},
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Asana_ListTags",
                            "arguments": '{"limit":5}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "count": 2,
                    "next_page": {
                        "has_more_results": True,
                        "next_page_token": "abc123",
                    },
                    "workspaces": [
                        {
                            "id": "1234567890",
                            "name": "Tag Hello",
                        },
                        {
                            "id": "1234567891",
                            "name": "Tag World",
                        },
                    ],
                }),
                "tool_call_id": "call_1",
                "name": "Asana_ListTags",
            },
            {
                "role": "assistant",
                "content": "Here are two tags in Asana:\n\n1. Tag Hello\n2. Tag World",
            },
        ],
    )

    return suite


@tool_eval()
def create_tag_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="create tag eval suite",
        system_message="You are an AI assistant with access to Asana tools. Use them to help the user with their tasks.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Create tag",
        user_message="Create a 'Hello World' tag in Asana.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_tag,
                args={
                    "name": "Hello World",
                    "description": None,
                    "color": None,
                    "workspace_id": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=0.7),
            BinaryCritic(critic_field="description", weight=0.1),
            BinaryCritic(critic_field="color", weight=0.1),
            BinaryCritic(critic_field="workspace_id", weight=0.1),
        ],
    )

    suite.add_case(
        name="Create tag with description and color",
        user_message="Create a dark orange tag 'Attention' in Asana with the description 'This is a tag for important tasks'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_tag,
                args={
                    "name": "Attention",
                    "description": "This is a tag for important tasks",
                    "color": "dark-orange",
                    "workspace_id": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=0.3),
            BinaryCritic(critic_field="description", weight=0.3),
            BinaryCritic(critic_field="color", weight=0.3),
            BinaryCritic(critic_field="workspace_id", weight=0.1),
        ],
    )

    suite.add_case(
        name="Create tag in a specific workspace",
        user_message="Create a dark orange tag 'Attention' in Asana with the description 'This is a tag for important tasks' in the workspace '1234567890'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_tag,
                args={
                    "name": "Attention",
                    "description": "This is a tag for important tasks",
                    "color": "dark-orange",
                    "workspace_id": "1234567890",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=0.25),
            BinaryCritic(critic_field="description", weight=0.25),
            BinaryCritic(critic_field="color", weight=0.25),
            BinaryCritic(critic_field="workspace_id", weight=0.25),
        ],
    )

    return suite



def convert_document_to_html(document: dict) -> str:
    html = (
        "<html><head>"
        f"<title>{document['title']}</title>"
        f'<meta name="documentId" content="{document["documentId"]}">'
        "</head><body>"
    )
    for element in document["body"]["content"]:
        html += convert_structural_element(element)
    html += "</body></html>"
    return html


def convert_structural_element(element: dict, wrap_paragraphs: bool = True) -> str:
    if "sectionBreak" in element or "tableOfContents" in element:
        return ""

    elif "paragraph" in element:
        paragraph_content = ""

        prepend, append = get_paragraph_style_tags(
            style=element["paragraph"]["paragraphStyle"],
            wrap_paragraphs=wrap_paragraphs,
        )

        for item in element["paragraph"]["elements"]:
            if "textRun" not in item:
                continue
            paragraph_content += extract_paragraph_content(item["textRun"])

        if not paragraph_content:
            return ""

        return f"{prepend}{paragraph_content.strip()}{append}"

    elif "table" in element:
        table = [
            [
                "".join([
                    convert_structural_element(element=cell_element, wrap_paragraphs=False)
                    for cell_element in cell["content"]
                ])
                for cell in row["tableCells"]
            ]
            for row in element["table"]["tableRows"]
        ]
        return table_list_to_html(table)

    else:
        raise ValueError(f"Unknown document body element type: {element}")


def extract_paragraph_content(text_run: dict) -> str:
    content = text_run["content"]
    style = text_run["textStyle"]
    return apply_text_style(content, style)


def apply_text_style(content: str, style: dict) -> str:
    content = content.rstrip("\n")
    content = content.replace("\n", "<br>")
    italic = style.get("italic", False)
    bold = style.get("bold", False)
    if italic:
        content = f"<i>{content}</i>"
    if bold:
        content = f"<b>{content}</b>"
    return content


def get_paragraph_style_tags(style: dict, wrap_paragraphs: bool = True) -> tuple[str, str]:
    named_style = style["namedStyleType"]
    if named_style == "NORMAL_TEXT":
        return ("<p>", "</p>") if wrap_paragraphs else ("", "")
    elif named_style == "TITLE":
        return "<h1>", "</h1>"
    elif named_style == "SUBTITLE":
        return "<h2>", "</h2>"
    elif named_style.startswith("HEADING_"):
        try:
            heading_level = int(named_style.split("_")[1])
        except ValueError:
            return ("<p>", "</p>") if wrap_paragraphs else ("", "")
        else:
            return f"<h{heading_level}>", f"</h{heading_level}>"
    return ("<p>", "</p>") if wrap_paragraphs else ("", "")


def table_list_to_html(table: list[list[str]]) -> str:
    html = "<table>"
    for row in table:
        html += "<tr>"
        for cell in row:
            if cell.endswith("<br>"):
                cell = cell[:-4]
            html += f"<td>{cell}</td>"
        html += "</tr>"
    html += "</table>"
    return html



import re
from dataclasses import dataclass, field
from typing import Any

from bs4 import BeautifulSoup
from msgraph.generated.models.body_type import BodyType
from msgraph.generated.models.email_address import EmailAddress
from msgraph.generated.models.item_body import ItemBody
from msgraph.generated.models.message import Message as GraphMessage
from msgraph.generated.models.recipient import Recipient as GraphRecipient


@dataclass
class Recipient:
    """A recipient of an email message."""

    email_address: str = ""
    name: str = ""

    @classmethod
    def from_sdk(cls, recipient: GraphRecipient) -> "Recipient":
        """Convert a Microsoft Graph SDK Recipient object to a Recipient dataclass."""
        address = (
            recipient.email_address.address
            if recipient and recipient.email_address and recipient.email_address.address
            else ""
        )
        name = (
            recipient.email_address.name
            if recipient and recipient.email_address and recipient.email_address.name
            else ""
        )
        return cls(email_address=address, name=name)

    def to_dict(self) -> dict[str, str]:
        return {"email_address": self.email_address, "name": self.name}

    def to_sdk(self) -> GraphRecipient:
        """Converts the Recipient dataclass to a Microsoft Graph SDK Recipient object."""
        recipient = GraphRecipient()
        email_address = EmailAddress()
        email_address.address = self.email_address
        email_address.name = self.name
        recipient.email_address = email_address
        return recipient


@dataclass
class Message:
    """An email message in Outlook."""

    bcc_recipients: list[Recipient] = field(default_factory=list)
    cc_recipients: list[Recipient] = field(default_factory=list)
    reply_to: list[Recipient] = field(default_factory=list)
    to_recipients: list[Recipient] = field(default_factory=list)
    from_: Recipient = field(default_factory=Recipient)
    subject: str = ""
    body: str = ""
    conversation_id: str = ""
    conversation_index: str = ""
    flag: dict[str, str] = field(default_factory=dict)
    has_attachments: bool = False
    importance: str = ""
    is_read: bool = False
    received_date_time: str = ""
    web_link: str = ""
    is_draft: bool = True
    message_id: str = ""  # The unique identifier of the email message. Read-only.

    @staticmethod
    def _safe_str(value: Any) -> str:
        if not value:
            return ""
        if isinstance(value, bytes | bytearray):
            return value.decode("utf-8", errors="ignore")
        return str(value)

    @staticmethod
    def _safe_bool(value: Any) -> bool:
        return bool(value)

    @staticmethod
    def _parse_body(mime: str) -> str:
        if not mime:
            return ""
        soup = BeautifulSoup(mime, "html.parser")
        text = soup.get_text(separator=" ")
        # Replace multiple newlines with a single newline
        text = re.sub(r"\n+", "\n", text)
        # Replace multiple spaces with a single space
        text = re.sub(r"\s+", " ", text)
        # Remove leading/trailing whitespace from each line
        text = "\n".join(line.strip() for line in text.split("\n"))

        return text

    @staticmethod
    def _parse_importance(value: Any) -> str:
        return value.value if getattr(value, "value", None) else ""

    @staticmethod
    def _parse_flag(flag: Any) -> dict[str, str]:
        if not flag:
            return {"flag_status": "", "due_date_time": ""}
        status = flag.flag_status.value if getattr(flag, "flag_status", None) else ""
        due = ""
        if getattr(flag, "due_date_time", None) and getattr(flag.due_date_time, "date_time", None):
            due = Message._safe_str(flag.due_date_time.date_time)
        return {"flag_status": status, "due_date_time": due}

    @classmethod
    def from_sdk(cls, msg: GraphMessage) -> "Message":
        """Convert a Microsoft Graph SDK Message object to a Message dataclass."""
        text = cls._parse_body(msg.body.content if msg.body and msg.body.content else "")
        return cls(
            bcc_recipients=[
                Recipient.from_sdk(recipient) for recipient in msg.bcc_recipients or []
            ],
            cc_recipients=[Recipient.from_sdk(recipient) for recipient in msg.cc_recipients or []],
            reply_to=[Recipient.from_sdk(recipient) for recipient in msg.reply_to or []],
            to_recipients=[Recipient.from_sdk(recipient) for recipient in msg.to_recipients or []],
            from_=Recipient.from_sdk(msg.from_) if msg.from_ else Recipient(),
            subject=cls._safe_str(msg.subject),
            body=text,
            conversation_id=cls._safe_str(msg.conversation_id),
            conversation_index=(
                msg.conversation_index.decode("utf-8", errors="ignore")
                if isinstance(msg.conversation_index, bytes | bytearray)
                else cls._safe_str(msg.conversation_index)
            ),
            flag=cls._parse_flag(msg.flag),
            has_attachments=cls._safe_bool(msg.has_attachments),
            importance=cls._parse_importance(msg.importance),
            is_read=cls._safe_bool(msg.is_read),
            received_date_time=(
                msg.received_date_time.isoformat() if msg.received_date_time else ""
            ),
            web_link=cls._safe_str(msg.web_link),
            is_draft=cls._safe_bool(msg.is_draft),
            message_id=cls._safe_str(msg.id),
        )

    def to_sdk(self) -> GraphMessage:
        """Converts the Message dataclass to a Microsoft Graph SDK Message object."""
        sdk_msg = GraphMessage()
        sdk_msg.subject = self.subject
        body_obj = ItemBody()
        body_obj.content = self.body
        body_obj.content_type = BodyType.Text
        sdk_msg.body = body_obj
        sdk_msg.is_draft = self.is_draft
        sdk_msg.to_recipients = [r.to_sdk() for r in self.to_recipients]
        sdk_msg.cc_recipients = [r.to_sdk() for r in self.cc_recipients]
        sdk_msg.bcc_recipients = [r.to_sdk() for r in self.bcc_recipients]
        sdk_msg.reply_to = [r.to_sdk() for r in self.reply_to]

        return sdk_msg

    def to_dict(self) -> dict[str, Any]:
        """Converts the Message dataclass to a dictionary."""
        return {
            "bcc_recipients": [recipient.to_dict() for recipient in self.bcc_recipients],
            "cc_recipients": [recipient.to_dict() for recipient in self.cc_recipients],
            "reply_to": [recipient.to_dict() for recipient in self.reply_to],
            "to_recipients": [recipient.to_dict() for recipient in self.to_recipients],
            "from": self.from_.to_dict(),
            "subject": self.subject,
            "body": self.body,
            "conversation_id": self.conversation_id,
            "conversation_index": self.conversation_index,
            "flag": self.flag,
            "has_attachments": self.has_attachments,
            "importance": self.importance,
            "is_read": self.is_read,
            "received_date_time": self.received_date_time,
            "web_link": self.web_link,
            "is_draft": self.is_draft,
            "message_id": self.message_id,
        }

    def update_recipient_lists(
        self,
        to_add: list[str] | None = None,
        to_remove: list[str] | None = None,
        cc_add: list[str] | None = None,
        cc_remove: list[str] | None = None,
        bcc_add: list[str] | None = None,
        bcc_remove: list[str] | None = None,
    ) -> None:
        """Update each recipient list of the message.

        This function updates the recipient lists of the message by first adding new recipients
        and then removing existing recipients. Therefore, if an email address is both
        added and removed, then it will not be included in the returned list.
        """
        for attr, add_emails_input, remove_emails_input in (
            ("to_recipients", to_add, to_remove),
            ("cc_recipients", cc_add, cc_remove),
            ("bcc_recipients", bcc_add, bcc_remove),
        ):
            current_recipients = getattr(self, attr) or []
            # Add recipients
            existing_emails = {r.email_address.lower() for r in current_recipients}
            new_additions = [
                Recipient(email_address=email)
                for email in (add_emails_input or [])
                if email.lower() not in existing_emails
            ]
            # Remove recipients
            updated_list = current_recipients + new_additions
            remove_emails = {email.lower() for email in (remove_emails_input or [])}
            updated_list = [
                recipient
                for recipient in updated_list
                if recipient.email_address.lower() not in remove_emails
            ]
            # Update the message's attribute with the new list
            setattr(self, attr, updated_list)



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Microsoft
from msgraph.generated.users.item.calendar.calendar_view.calendar_view_request_builder import (
    CalendarViewRequestBuilder,
)

from arcade_microsoft.client import get_client
from arcade_microsoft.outlook_calendar._utils import (
    convert_timezone_to_offset,
    create_timezone_request_config,
    get_default_calendar_timezone,
    replace_timezone_offset,
    validate_date_times,
)
from arcade_microsoft.outlook_calendar.models import Event


@tool(requires_auth=Microsoft(scopes=["MailboxSettings.Read", "Calendars.ReadBasic"]))
async def list_events_in_time_range(
    context: ToolContext,
    start_date_time: Annotated[
        str,
        "The start date and time of the time range, represented in "
        "ISO 8601 format. Timezone offset is ignored. For example, 2025-04-24T19:00:00",
    ],
    end_date_time: Annotated[
        str,
        "The end date and time of the time range, represented in "
        "ISO 8601 format. Timezone offset is ignored. For example, 2025-04-24T19:30:00",
    ],
    limit: Annotated[int, "The maximum number of events to return. Max 1000. Defaults to 10"] = 10,
) -> Annotated[dict, "A dictionary containing a list of events"]:
    """List events in the user's calendar in a specific time range.

    Ignores timezone offsets provided in the start_date_time and end_date_time parameters.
    Instead, uses the user's default calendar timezone to filter events.
    If the user has not set a timezone for their calendar, then the timezone will be UTC.
    """
    # Validate inputs
    validate_date_times(start_date_time, end_date_time)

    client = get_client(context.get_auth_token_or_empty())
    time_zone = await get_default_calendar_timezone(client)
    time_zone_offset = convert_timezone_to_offset(time_zone)
    start_date_time = replace_timezone_offset(start_date_time, time_zone_offset)
    end_date_time = replace_timezone_offset(end_date_time, time_zone_offset)
    query_params = CalendarViewRequestBuilder.CalendarViewRequestBuilderGetQueryParameters(
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        top=max(1, min(limit, 1000)),
    )
    request_config = create_timezone_request_config(time_zone, query_params)

    response = await client.me.calendar.calendar_view.get(request_config)
    events = [Event.from_sdk(event).to_dict() for event in response.value]  # type: ignore[union-attr]

    return {"events": events, "num_events": len(events)}



from arcade_jira.tools.attachments import (
    attach_file_to_issue,
    download_attachment,
    get_attachment_metadata,
    list_issue_attachments_metadata,
)
from arcade_jira.tools.comments import (
    add_comment_to_issue,
    get_comment_by_id,
    get_issue_comments,
)
from arcade_jira.tools.issues import (
    add_labels_to_issue,
    create_issue,
    get_issue_by_id,
    get_issue_type_by_id,
    get_issues_without_id,
    list_issue_types_by_project,
    remove_labels_from_issue,
    search_issues_with_jql,
    update_issue,
)
from arcade_jira.tools.labels import list_labels
from arcade_jira.tools.priorities import (
    get_priority_by_id,
    list_priorities_associated_with_a_priority_scheme,
    list_priorities_available_to_a_project,
    list_priorities_available_to_an_issue,
    list_priority_schemes,
    list_projects_associated_with_a_priority_scheme,
)
from arcade_jira.tools.projects import get_project_by_id, search_projects
from arcade_jira.tools.transitions import (
    get_transition_by_status_name,
    get_transitions_available_for_issue,
    transition_issue_to_new_status,
)
from arcade_jira.tools.users import get_user_by_id, get_users_without_id, list_users

__all__ = [
    # Attachments tools
    "attach_file_to_issue",
    "download_attachment",
    "get_attachment_metadata",
    "list_issue_attachments_metadata",
    # Comments tools
    "add_comment_to_issue",
    "get_comment_by_id",
    "get_issue_comments",
    # Issues tools
    "add_labels_to_issue",
    "create_issue",
    "get_issue_by_id",
    "get_issue_type_by_id",
    "get_issues_without_id",
    "list_issue_types_by_project",
    "remove_labels_from_issue",
    "search_issues_with_jql",
    "update_issue",
    # Labels tools
    "list_labels",
    # Priorities tools
    "get_priority_by_id",
    "list_priority_schemes",
    "list_priorities_associated_with_a_priority_scheme",
    "list_projects_associated_with_a_priority_scheme",
    "list_priorities_available_to_a_project",
    "list_priorities_available_to_an_issue",
    # Projects tools
    "get_project_by_id",
    "search_projects",
    # Transitions tools
    "get_transition_by_status_name",
    "get_transitions_available_for_issue",
    "transition_issue_to_new_status",
    # Users tools
    "get_user_by_id",
    "get_users_without_id",
    "list_users",
]



from arcade_gmail.tools.gmail import (
    change_email_labels,
    create_label,
    delete_draft_email,
    get_thread,
    list_draft_emails,
    list_emails,
    list_emails_by_header,
    list_labels,
    list_threads,
    reply_to_email,
    search_threads,
    send_draft_email,
    send_email,
    trash_email,
    update_draft_email,
    write_draft_email,
    write_draft_reply_email,
)

__all__ = [
    "change_email_labels",
    "create_label",
    "delete_draft_email",
    "get_thread",
    "list_draft_emails",
    "list_emails",
    "list_emails_by_header",
    "list_labels",
    "list_threads",
    "reply_to_email",
    "search_threads",
    "send_draft_email",
    "send_email",
    "trash_email",
    "update_draft_email",
    "write_draft_email",
    "write_draft_reply_email",
]



import string
from collections.abc import Callable
from typing import Any, cast

from arcade_salesforce.constants import ASSOCIATION_REFERENCE_FIELDS, GLOBALLY_IGNORED_FIELDS
from arcade_salesforce.enums import SalesforceObject


def remove_fields_globally_ignored(clean_func: Callable[[dict], dict]) -> Callable[[dict], dict]:
    """Remove fields that are irrelevant for our tools' responses.

    The main purpose of cleaning and removing unnecessary fields from data is to make our
    responses more concise and save on LLM tokens/context window.
    """

    def global_cleaner(data: dict) -> dict:
        cleaned_data = {}
        for key, value in data.items():
            if key in GLOBALLY_IGNORED_FIELDS or value is None:
                continue

            if isinstance(value, dict):
                cleaned_data[key] = global_cleaner(value)

            elif isinstance(value, list | tuple | set):
                cleaned_items = []
                for item in value:
                    if isinstance(item, dict):
                        cleaned_items.append(global_cleaner(item))
                    else:
                        cleaned_items.append(item)
                cleaned_data[key] = cleaned_items  # type: ignore[assignment]
            else:
                cleaned_data[key] = value
        return cleaned_data

    def wrapper(data: dict) -> dict:
        return global_cleaner(clean_func(data))

    return wrapper


def clean_object_data(data: dict) -> dict:
    """Clean Salesforce object dictionary based on the object type."""

    obj_type = data["attributes"]["type"]
    if obj_type == SalesforceObject.ACCOUNT.value:
        return clean_account_data(data)
    elif obj_type == SalesforceObject.CONTACT.value:
        return clean_contact_data(data)
    elif obj_type == SalesforceObject.LEAD.value:
        return clean_lead_data(data)
    elif obj_type == SalesforceObject.NOTE.value:
        return clean_note_data(data)
    elif obj_type == SalesforceObject.TASK.value:
        return clean_task_data(data)
    elif obj_type == SalesforceObject.USER.value:
        return clean_user_data(data)
    raise ValueError(f"Unknown object type: '{obj_type}' in object: {data}")


@remove_fields_globally_ignored
def clean_account_data(data: dict) -> dict:
    data["AccountType"] = data["Type"]
    del data["Type"]
    data["ObjectType"] = SalesforceObject.ACCOUNT.value
    ignore_fields = [
        "BillingCity",
        "BillingCountry",
        "BillingCountryCode",
        "BillingPostalCode",
        "BillingState",
        "BillingStateCode",
        "BillingStreet",
        "LastActivityDate",
        "ShippingCity",
        "ShippingCountry",
        "ShippingCountryCode",
        "ShippingPostalCode",
        "ShippingState",
        "ShippingStateCode",
        "ShippingStreet",
    ]
    return {k: v for k, v in data.items() if v is not None and k not in ignore_fields}


@remove_fields_globally_ignored
def clean_contact_data(data: dict) -> dict:
    data["ObjectType"] = SalesforceObject.CONTACT.value
    ignore_fields = ["IsEmailBounced", "IsPriorityRecord"]
    return {k: v for k, v in data.items() if v is not None and k not in ignore_fields}


@remove_fields_globally_ignored
def clean_lead_data(data: dict) -> dict:
    data["ObjectType"] = SalesforceObject.LEAD.value
    return data


@remove_fields_globally_ignored
def clean_opportunity_data(data: dict) -> dict:
    data["ObjectType"] = SalesforceObject.OPPORTUNITY.value
    data["Amount"] = {
        "Value": data["Amount"],
        "ClosingProbability": data["Probability"]
        if not isinstance(data["Probability"], int | float)
        else data["Probability"] / 100,
        "ExpectedRevenue": data["ExpectedRevenue"],
    }
    del data["Probability"]
    del data["ExpectedRevenue"]
    return data


@remove_fields_globally_ignored
def clean_note_data(data: dict) -> dict:
    data["ObjectType"] = SalesforceObject.NOTE.value
    return data


@remove_fields_globally_ignored
def clean_task_data(data: dict) -> dict:
    data["ObjectType"] = data["TaskSubtype"]
    data["AssociatedToWhom"] = data["WhoId"]
    ignore_fields = [
        "IsArchived",
        "IsClosed",
        "IsDeleted",
        "IsHighPriority",
        "IsRecurrence",
        "IsReminderSet",
        "TaskSubtype",
        "WhoId",
    ]

    if data["ObjectType"] == SalesforceObject.EMAIL.value:
        data["Email"] = format_email(data["Description"])
        del data["ActivityDate"]
        del data["CompletedDateTime"]
        del data["Description"]
        del data["Subject"]
        del data["OwnerId"]
        del data["Priority"]
        del data["Status"]

    return {k: v for k, v in data.items() if v is not None and k not in ignore_fields}


@remove_fields_globally_ignored
def clean_user_data(data: dict) -> dict:
    data["ObjectType"] = SalesforceObject.USER.value
    ignore_fields = [
        "attributes",
        "CleanStatus",
        "LastReferencedDate",
        "LastViewedDate",
        "SystemModstamp",
    ]
    return {k: v for k, v in data.items() if v is not None and k not in ignore_fields}


def get_ids_referenced(*data_objects: dict) -> set[str]:
    """Build a set of IDs referenced in a list of dictionaries.

    Args:
        data_objects: The list of dictionaries to search for referenced IDs.

    Returns:
        The set of IDs referenced in the dictionaries.
    """
    referenced_ids = set()
    for data in data_objects:
        for field in ASSOCIATION_REFERENCE_FIELDS:
            if field in data:
                referenced_ids.add(data[field])
    return referenced_ids


def expand_associations(data: dict, objects_by_id: dict) -> dict:
    """Expand a Salesforce object referenced by ID with metadata about the object.

    Args:
        data: The dictionary to expand.
        objects_by_id: The dictionary of objects metadata by ID.

    Returns:
        The expanded dictionary.
    """
    for field in ASSOCIATION_REFERENCE_FIELDS:
        if field not in data:
            continue

        associated_object = objects_by_id.get(data[field])

        if isinstance(associated_object, dict):
            del data[field]
            data[field.rstrip("Id")] = simplified_object_data(associated_object)

    return data


def simplified_object_data(data: dict) -> dict:
    return {
        "Id": data["Id"],
        "Name": data["Name"],
        "ObjectType": get_object_type(data),
    }


def get_object_type(data: dict) -> str:
    obj_type = data.get("ObjectType") or data["attributes"]["type"]
    return cast(str, obj_type)


def build_soql_query(query: str, **kwargs: Any) -> str:
    """Build a SOQL query with sanitized arguments.

    Args:
        query: The SOQL query to build.
        **kwargs: The arguments to sanitize and format into the query.

    Returns:
        The SOQL query with the arguments formatted.
    """
    return query.format(**{key: sanitize_soql_argument(value) for key, value in kwargs.items()})


def sanitize_soql_argument(value: Any) -> str:
    """Sanitize an argument for a SOQL query.

    The goal is to prevent SQL injection, since Salesforce does not support parameterized queries.
    This function will only accept ASCII letters and digits in the value.

    Args:
        value: The value to sanitize.

    Returns:
        The sanitized value.
    """
    allowed_chars = string.ascii_letters + string.digits
    if not isinstance(value, str):
        value = str(value)
    return "".join([char for char in value if char in allowed_chars])


def format_email(description: str) -> dict:
    """Turns a Salesforce email description string into a more readable dictionary.

    Args:
        description: The email description to format.

    Returns:
        The formatted email as a dictionary.
    """
    email = {
        "To": description.split("To:")[1].split("\n")[0].strip(),
        "CC": description.split("CC:")[1].split("\n")[0].strip(),
        "BCC": description.split("BCC:")[1].split("\n")[0].strip(),
        "Attachment": description.split("Attachment:")[1].split("\n")[0].strip(),
        "Subject": description.split("Subject:")[1].split("\n")[0].strip(),
        "Body": description.split("Body:\n")[1].strip(),
    }

    if email["Attachment"] == "--none--":
        del email["Attachment"]

    return email


def remove_none_values(data: dict) -> dict:
    """Remove None values from a dictionary.

    Args:
        data: The dictionary to clean.

    Returns:
        The cleaned dictionary.
    """
    return {k: v for k, v in data.items() if v is not None}



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_reddit
from arcade_reddit.tools import (
    submit_text_post,
)
from arcade_reddit.tools.submit import comment_on_post
from evals.additional_messages import get_post_in_subreddit_messages
from evals.critics import AnyOfCritic

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_reddit)


@tool_eval()
def reddit_submit_text_post_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_submit_text_post_1",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="reddit_submit_text_post_1",
        user_message=(
            "Post this in AskReddit - 'Why is the sky blue?'. I dont want replies sent to my dms"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=submit_text_post,
                args={
                    "subreddit": "AskReddit",
                    "title": "Why is the sky blue?",
                    "send_replies": False,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="subreddit", weight=0.3),
            BinaryCritic(critic_field="title", weight=0.3),
            BinaryCritic(critic_field="body", weight=0.3),
            BinaryCritic(critic_field="send_replies", weight=0.1),
        ],
    )

    return suite


@tool_eval()
def reddit_comment_on_post_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="reddit_comment_on_post_1",
        system_message=(
            "You are an AI assistant with access to reddit tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    comment = "Your dog's four-legged structure is a manifestation of tetrapodal evolution, where natural selection has optimized limb development for biomechanical stability and efficient terrestrial locomotion. Duh!"  # noqa: E501
    suite.add_case(
        name="reddit_comment_on_post_1",
        user_message=(f"here's my comment - {comment}"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=comment_on_post,
                args={
                    "post_identifier": [  # post_identifier can be any of the following
                        "1abcdef",
                        "t3_1abcdef",
                        "https://www.reddit.com/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                        "/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/",
                    ],
                    "text": comment,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            AnyOfCritic(critic_field="post_identifier", weight=0.5),
            SimilarityCritic(critic_field="text", weight=0.5),
        ],
        additional_messages=get_post_in_subreddit_messages,
    )

    return suite



from dataclasses import dataclass
from typing import Any

from arcade_evals.critic import Critic


@dataclass
class AnyOfCritic(Critic):
    """
    A critic that checks if the actual value matches any of the expected values.
    In other words, it checks if the actual value is in the expected list.
    """

    def evaluate(self, expected: list[Any], actual: Any) -> dict[str, float | bool]:
        match = actual in expected
        return {"match": match, "score": self.weight if match else 0.0}


@dataclass
class ListCritic(Critic):
    """
    A critic for comparing two lists.
    """

    def __init__(
        self,
        critic_field: str,
        weight: float = 1.0,
        order_matters: bool = True,
        duplicates_matter: bool = True,
    ):
        self.critic_field = critic_field
        self.weight = weight
        self.order_matters = order_matters
        self.duplicates_matter = duplicates_matter

    def evaluate(self, expected: list[Any], actual: list[Any]) -> dict[str, float | bool]:
        match = actual == expected if self.order_matters else set(actual) == set(expected)
        if self.duplicates_matter:
            match = match and len(actual) == len(expected)

        return {"match": match, "score": self.weight if match else 0.0}



import pytest


class DummyContext:
    def get_secret(self, key: str) -> str | None:
        if key.lower() == "serp_api_key":
            return "dummy_key"
        return None


@pytest.fixture
def dummy_context():
    return DummyContext()



from datetime import timedelta

from arcade_evals import (
    BinaryCritic,
    DatetimeCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_google_calendar
from arcade_google_calendar.enums import EventVisibility, SendUpdatesOptions
from arcade_google_calendar.tools import (
    create_event,
    delete_event,
    list_calendars,
    list_events,
    update_event,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_module(arcade_google_calendar)

history_after_list_events = [
    {"role": "user", "content": "do i have any events on my calendar for today?"},
    {
        "role": "assistant",
        "content": "Please go to this URL and authorize the action: \n[Link](https://accounts.google.com/o/oauth2/v2/auth?)",
    },
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_uHdRlr4z7sFm39ZrPsE5wcdT",
                "type": "function",
                "function": {
                    "name": "GoogleCalendar_ListEvents",
                    "arguments": '{"min_end_datetime":"2024-09-26T00:00:00-07:00","max_start_datetime":"2024-09-27T00:00:00-07:00"}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"events_count": 3, "events": [{"creator": {"email": "john@example.com", "self": true}, "description": "1:1 meeting with Joe", "end": {"dateTime": "2024-09-26T00:15:00-07:00", "timeZone": "America/Los_Angeles"}, "eventType": "default", "htmlLink": "https://www.google.com/calendar/event?eid=01234", "id": "10009199283838467", "location": "622 Rainbow Ave, South San Francisco, CA 94080, USA", "organizer": {"email": "john@example.com", "self": true}, "start": {"dateTime": "2024-09-25T23:15:00-07:00", "timeZone": "America/Los_Angeles"}, "summary": "1:1 meeting"}, {"attendees": [{"email": "joe@example.com", "responseStatus": "accepted"}], "creator": {"email": "john@example.com", "self": true}, "description": "This is just a test", "end": {"dateTime": "2024-09-26T14:00:00-07:00", "timeZone": "America/Los_Angeles"}, "eventType": "default", "htmlLink": "https://www.google.com/calendar/event?eid=OXB2OGFwcmZraWk1N234324", "id": "00099992228181818181", "organizer": {"email": "john@example.com", "self": true}, "start": {"dateTime": "2024-09-26T12:00:00-07:00", "timeZone": "America/Los_Angeles"}, "summary": "API test"}, {"attendees": [{"email": "henry@example.com", "responseStatus": "needsAction"}], "creator": {"email": "john@example.com", "self": true}, "end": {"dateTime": "2024-09-26T17:00:00-07:00", "timeZone": "America/Los_Angeles"}, "eventType": "default", "htmlLink": "https://www.google.com/calendar/event?eid=Z3I1ZzE4b324534556", "id": "gr5g18lf88tfpp3vkareukkc7g", "location": "611 Rainbow Road", "organizer": {"email": "john@example.com", "self": true}, "start": {"dateTime": "2024-09-26T15:00:00-07:00", "timeZone": "America/Los_Angeles"}, "summary": "Focus Time", "visibility": "public"}]}',
        "tool_call_id": "call_uHdRlr4z7sFm39ZrPsE5wcdT",
        "name": "GoogleCalendar_ListEvents",
    },
    {
        "role": "assistant",
        "content": "Yes, you have three events on your calendar for today:\n\n1. **Event:** Test2\n   - **Time:** 23:15 - 00:15 (PST)\n   - **Location:** 611 Gateway Blvd, South San Francisco, CA 94080, USA\n   - **Description:** 1:1 meeting with Joe\n  2. **Event:** API Test\n   - **Time:** 12:00 - 14:00 (PST)\n  **Description:** This is just a test\n   - [View Event](https://www.google.com/calendar/event?eid=OXB2OGFwcmZraWk1NGUwa24xaTNya2lvZDggZXJpY0BhcmNhZGUtYWkuY29t)\n\n3. **Event:** Focus Time\n   - **Time:** 15:00 - 17:00 (PST)\n   - **Location:** 611 Rainbow Road\n   - **Visibility:** Public\n   - [View Event](https://www.google.com/calendar/event?eid=Z3I1ZzE4bGY4OHRmcHAzdmthcmV1a2tjN2cgZXJpY0BhcmNhZGUtYWkuY29t)\n\nIf you need more details or help with anything else, feel free to ask!",
    },
]


@tool_eval()
def calendar_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Calendar tools."""
    suite = EvalSuite(
        name="Calendar Tools Evaluation",
        system_message=(
            "You are an AI assistant that can create, list, update, and delete events using the provided tools. Today is 2024-09-26"
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Cases for list_calendars
    suite.add_case(
        name="List Calendars",
        user_message=("What calendars do I have?"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_calendars,
                args={},
            )
        ],
        critics=[],
    )

    # Cases for create_event
    suite.add_case(
        name="Create calendar event",
        user_message=(
            "Create a meeting for 'Team Meeting' starting on September 26, 2024, from 11:45pm to 12:15am. Invite johndoe@example.com"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_event,
                args={
                    "summary": "Team Meeting",
                    "start_datetime": "2024-09-26T23:45:00",
                    "end_datetime": "2024-09-27T00:15:00",
                    "calendar_id": "primary",
                    "attendee_emails": ["johndoe@example.com"],
                    "visibility": EventVisibility.DEFAULT,
                    "description": "Team Meeting",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="summary", weight=0.2),
            DatetimeCritic(
                critic_field="start_datetime", weight=0.2, tolerance=timedelta(seconds=10)
            ),
            DatetimeCritic(
                critic_field="end_datetime", weight=0.2, tolerance=timedelta(seconds=10)
            ),
            BinaryCritic(critic_field="attendee_emails", weight=0.2),
            BinaryCritic(critic_field="description", weight=0.1),
            BinaryCritic(critic_field="location", weight=0.1),
        ],
    )

    # Cases for list_events
    suite.add_case(
        name="List calendar events",
        user_message="Do I have any events on my calendar today?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_events,
                args={
                    "min_end_datetime": "2024-09-26T00:00:00",
                    "max_start_datetime": "2024-09-27T00:00:00",
                    "calendar_id": "primary",
                    "max_results": 10,
                },
            )
        ],
        critics=[
            DatetimeCritic(
                critic_field="min_end_datetime", weight=0.3, tolerance=timedelta(hours=1)
            ),
            DatetimeCritic(
                critic_field="max_start_datetime", weight=0.3, tolerance=timedelta(hours=1)
            ),
            BinaryCritic(critic_field="calendar_id", weight=0.2),
            BinaryCritic(critic_field="max_results", weight=0.2),
        ],
    )

    # Cases for update_event
    suite.add_case(
        name="Update a calendar event",
        user_message=(
            "Oh no! I can't make it to the API Test since I have lunch with an old friend at that time. "
            "Change my meeting tomorrow at 3pm to 4pm. Let everyone know."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=update_event,
                args={
                    "event_id": "00099992228181818181",
                    "updated_start_datetime": "2024-09-27T16:00:00",
                    "updated_end_datetime": "2024-09-27T18:00:00",
                    "updated_calendar_id": "primary",
                    "updated_summary": "API Test",
                    "updated_description": "API Test",
                    "updated_location": "611 Gateway Blvd",
                    "updated_visibility": EventVisibility.DEFAULT,
                    "attendee_emails_to_add": None,
                    "attendee_emails_to_remove": None,
                    "send_updates": SendUpdatesOptions.ALL,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="event_id", weight=0.4),
            DatetimeCritic(
                critic_field="updated_start_datetime", weight=0.2, tolerance=timedelta(minutes=15)
            ),
            DatetimeCritic(
                critic_field="updated_end_datetime",
                weight=0.2,
                tolerance=timedelta(minutes=15),
            ),
            BinaryCritic(critic_field="send_updates", weight=0.2),
        ],
        additional_messages=history_after_list_events,
    )

    # Cases for delete_event
    suite.add_case(
        name="Delete a calendar event",
        user_message=(
            "I don't need to have focus time today. Please delete it from my calendar. Don't send any notifications."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=delete_event,
                args={
                    "event_id": "gr5g18lf88tfpp3vkareukkc7g",
                    "calendar_id": "primary",
                    "send_updates": SendUpdatesOptions.NONE,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="event_id", weight=0.6),
            BinaryCritic(critic_field="calendar_id", weight=0.2),
            BinaryCritic(critic_field="send_updates", weight=0.2),
        ],
        additional_messages=history_after_list_events,
    )

    return suite



import functools
import inspect
from typing import Any, Callable, TypeVar, Union

from arcade_tdk.auth import ToolAuthorization
from arcade_tdk.errors import ToolExecutionError
from arcade_tdk.utils import snake_to_pascal_case

T = TypeVar("T")


def tool(
    func: Callable | None = None,
    desc: str | None = None,
    name: str | None = None,
    requires_auth: Union[ToolAuthorization, None] = None,
    requires_secrets: Union[list[str], None] = None,
    requires_metadata: Union[list[str], None] = None,
) -> Callable:
    def decorator(func: Callable) -> Callable:
        func_name = str(getattr(func, "__name__", None))
        tool_name = name or snake_to_pascal_case(func_name)

        func.__tool_name__ = tool_name  # type: ignore[attr-defined]
        func.__tool_description__ = desc or inspect.cleandoc(func.__doc__ or "")  # type: ignore[attr-defined]
        func.__tool_requires_auth__ = requires_auth  # type: ignore[attr-defined]
        func.__tool_requires_secrets__ = requires_secrets  # type: ignore[attr-defined]
        func.__tool_requires_metadata__ = requires_metadata  # type: ignore[attr-defined]

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def func_with_error_handling(*args: Any, **kwargs: Any) -> Any:
                try:
                    return await func(*args, **kwargs)

                # make sure developer raised ToolExecutionError is not
                # reraised incorrectly.
                except ToolExecutionError:
                    raise
                except Exception as e:
                    raise ToolExecutionError(
                        message=f"Error in execution of {tool_name}",
                        developer_message=f"Error in {func_name}: {e!s}",
                    ) from e

        else:

            @functools.wraps(func)
            def func_with_error_handling(*args: Any, **kwargs: Any) -> Any:
                try:
                    return func(*args, **kwargs)
                except ToolExecutionError:
                    raise
                except Exception as e:
                    raise ToolExecutionError(
                        message=f"Error in execution of {tool_name}",
                        developer_message=f"Error in {func_name}: {e!s}",
                    ) from e

        return func_with_error_handling

    if func:
        return decorator(func)
    return decorator


def _tool_deprecated(message: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        func.__tool_deprecation_message__ = message  # type: ignore[attr-defined]
        return func

    return decorator


tool.deprecated = _tool_deprecated  # type: ignore[attr-defined]



from unittest.mock import AsyncMock, MagicMock

import pytest
from arcade_tdk.errors import ToolExecutionError

from arcade_linkedin.tools.share import create_text_post


@pytest.mark.asyncio
async def test_create_text_post_success(tool_context, mock_httpx_client):
    """Test successful creation of a LinkedIn text post."""
    # Mock response for a successful post creation
    mock_response = MagicMock()
    mock_response.status_code = 201
    mock_response.json.return_value = {"id": "1234567890"}
    # Ensure the mock is awaited properly
    mock_httpx_client.request = AsyncMock(return_value=mock_response)

    post_text = "Hello, LinkedIn!"
    result = await create_text_post(tool_context, post_text)

    expected_url = "https://www.linkedin.com/feed/update/1234567890/"
    assert result == expected_url
    mock_httpx_client.request.assert_called_once()


@pytest.mark.asyncio
async def test_create_text_post_no_user_id(tool_context):
    """Test error when user ID is not found in the context."""
    # Simulate missing user ID in the context
    tool_context.authorization.user_info = {}

    post_text = "Hello, LinkedIn!"
    with pytest.raises(ToolExecutionError, match="User ID not found"):
        await create_text_post(tool_context, post_text)



from enum import Enum


class HubspotObject(Enum):
    ACCOUNT = "account"
    CALL = "call"
    COMMUNICATION = "communication"
    COMPANY = "company"
    CONTACT = "contact"
    DEAL = "deal"
    EMAIL = "email"
    MEETING = "meeting"
    NOTE = "note"
    TASK = "task"

    @property
    def plural(self) -> str:
        if self.value == "company":
            return "companies"
        return f"{self.value}s"



import json

from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_evals.critic import BinaryCritic
from arcade_tdk import ToolCatalog

import arcade_jira
from arcade_jira.critics import CaseInsensitiveBinaryCritic, HasSubstringCritic
from arcade_jira.tools.issues import (
    get_issue_by_id,
    get_issues_without_id,
    list_issues,
    search_issues_with_jql,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_jira)


@tool_eval()
def get_issue_by_id_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Get issue by ID eval suite",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get issue by ID",
        user_message="Get the issue with ID '10000'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issue_by_id,
                args={
                    "issue_id": "10000",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="issue_id", weight=1.0),
        ],
    )

    suite.add_case(
        name="Get issue by Key",
        user_message="Get the issue ENG-103.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issue_by_id,
                args={
                    "issue_id": "ENG-103",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="issue_id", weight=1.0),
        ],
    )

    return suite


@tool_eval()
def get_issues_without_id_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Get issues without an ID",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks. "
            "Today is 2025-05-27 (Tuesday)."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get issues by keywords",
        user_message="Find the issue about implementing the message queue.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "keywords": "message queue",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            HasSubstringCritic(critic_field="keywords", weight=1.0),
        ],
    )

    suite.add_case(
        name="Get issues by due date",
        user_message="Which issues are due this month?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "due_from": "2025-05-01",
                    "due_until": "2025-05-31",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="due_from", weight=0.5),
            BinaryCritic(critic_field="due_until", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get issues by assignee, due date, status, priority and issue type",
        user_message=(
            "Find task issues assigned to John Doe that are in progress, "
            "with high priority, and due until the end of this month"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "assignee": "John Doe",
                    "due_from": None,
                    "due_until": "2025-05-31",
                    "status": "in progress",
                    "priority": "high",
                    "issue_type": "task",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="due_from", weight=0.1),
            BinaryCritic(critic_field="due_until", weight=0.1),
            CaseInsensitiveBinaryCritic(critic_field="assignee", weight=0.2),
            CaseInsensitiveBinaryCritic(critic_field="status", weight=0.2),
            CaseInsensitiveBinaryCritic(critic_field="priority", weight=0.2),
            CaseInsensitiveBinaryCritic(critic_field="issue_type", weight=0.2),
        ],
    )

    suite.add_case(
        name="Get issues by label and project name",
        user_message="Find issues labeled with version 2 in the Engineering project",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "project": "Engineering",
                    "labels": ["version 2"],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=0.5),
            BinaryCritic(critic_field="labels", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get issues by parent issue",
        user_message="Get the children issues of ENG-123",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "parent_issue": "ENG-123",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="parent_issue", weight=1.0),
        ],
    )

    suite.add_case(
        name="Paginate issues in multiple chat turns",
        user_message="Get the next page of issues",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issues_without_id,
                args={
                    "assignee": "john doe",
                    "priority": "high",
                    "status": "in progress",
                    "issue_type": "task",
                    "limit": 2,
                    "offset": 4,
                    "next_page_token": "1234567890",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            CaseInsensitiveBinaryCritic(critic_field="assignee", weight=1 / 7),
            CaseInsensitiveBinaryCritic(critic_field="priority", weight=1 / 7),
            CaseInsensitiveBinaryCritic(critic_field="status", weight=1 / 7),
            CaseInsensitiveBinaryCritic(critic_field="issue_type", weight=1 / 7),
            BinaryCritic(critic_field="limit", weight=1 / 7),
            BinaryCritic(critic_field="offset", weight=1 / 7),
            BinaryCritic(critic_field="next_page_token", weight=1 / 7),
        ],
        additional_messages=[
            {
                "role": "user",
                "content": (
                    "Find 2 tasks assigned to John Doe that are in progress, with high priority"
                ),
            },
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Jira_GetIssuesWithoutId",
                            "arguments": json.dumps({
                                "assignee": "John Doe",
                                "priority": "high",
                                "status": "in progress",
                                "issue_type": "task",
                                "limit": 2,
                                "offset": 0,
                            }),
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "issues": [
                        {
                            "id": "10001",
                            "key": "ENG-101",
                            "summary": "Implement the message queue",
                            "assignee": {
                                "id": "10010",
                                "name": "John Doe",
                                "email": "john.doe@example.com",
                            },
                            "status": {
                                "id": "10020",
                                "name": "In Progress",
                            },
                            "priority": {
                                "id": "10030",
                                "name": "High",
                            },
                            "issue_type": {
                                "id": "10040",
                                "name": "Task",
                            },
                            "project": {
                                "id": "10050",
                                "key": "ENG",
                                "name": "Engineering",
                            },
                        },
                        {
                            "id": "10002",
                            "key": "ENG-102",
                            "summary": "Deploy the message queue system",
                            "assignee": {
                                "id": "10010",
                                "name": "John Doe",
                                "email": "john.doe@example.com",
                            },
                            "status": {
                                "id": "10020",
                                "name": "In Progress",
                            },
                            "priority": {
                                "id": "10030",
                                "name": "High",
                            },
                            "issue_type": {
                                "id": "10040",
                                "name": "Task",
                            },
                            "project": {
                                "id": "10050",
                                "key": "ENG",
                                "name": "Engineering",
                            },
                        },
                    ],
                    "pagination": {
                        "limit": 2,
                        "total_results": 2,
                        "next_page_token": "1234567890",
                    },
                }),
                "tool_call_id": "call_1",
                "name": "Jira_GetIssuesWithoutId",
            },
            {
                "role": "assistant",
                "content": (
                    "Here are two issues:\n\n"
                    "1. ENG-101: Implement the message queue\n"
                    "2. ENG-102: Deploy the message queue system"
                ),
            },
        ],
    )

    return suite


@tool_eval()
def search_issues_with_jql_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Search issues with JQL",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks. "
            "Today is 2025-05-27 (Tuesday)."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    jql_query_str = 'text ~ "message queue" AND dueDate <= 2025-05-31'

    suite.add_case(
        name="Search issues by keywords",
        user_message=f"Search for up to 10 issues using the JQL query: {jql_query_str}",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_issues_with_jql,
                args={
                    "jql": jql_query_str,
                    "limit": 10,
                    "offset": 0,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            HasSubstringCritic(critic_field="jql", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
            BinaryCritic(critic_field="offset", weight=1 / 3),
        ],
    )

    return suite


@tool_eval()
def list_issues_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="List issues eval suite",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get me any one issue in Jira",
        user_message="Get me one issue in Jira.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issues,
                args={
                    "limit": 1,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="next_page_token", weight=0.5),
        ],
    )

    suite.add_case(
        name="List 10 issues in Jira",
        user_message="List 10 issues in Jira.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issues,
                args={
                    "limit": 10,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.5),
            BinaryCritic(critic_field="next_page_token", weight=0.5),
        ],
    )

    suite.add_case(
        name="List 10 issues in the Arcade project",
        user_message="List 10 issues in the Arcade project.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issues,
                args={
                    "project": "Arcade",
                    "limit": 50,
                    "next_page_token": None,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            CaseInsensitiveBinaryCritic(critic_field="project", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
            BinaryCritic(critic_field="next_page_token", weight=1 / 3),
        ],
    )

    return suite



import random
import string
from collections.abc import Callable

import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.fixture
def mock_chat_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.tools.chat.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def mock_users_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.tools.users.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def mock_user_retrieval_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.user_retrieval.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def mock_conversation_retrieval_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.conversation_retrieval.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def mock_message_retrieval_slack_client(mocker):
    mock_client = mocker.patch("arcade_slack.message_retrieval.AsyncWebClient", autospec=True)
    return mock_client.return_value


@pytest.fixture
def random_str_factory():
    def random_str_factory(length: int = 10):
        return "".join(random.choices(string.ascii_letters + string.digits, k=length))  # noqa: S311

    return random_str_factory


@pytest.fixture
def random_ts_factory():
    def random_ts_factory():
        return f"{random.uniform(1735689600.000000, 1751327999.999999)}"  # noqa: S311

    return random_ts_factory


@pytest.fixture
def dummy_channel_factory(random_str_factory: Callable[[int], str]):
    def dummy_channel_factory(
        id_: str | None = None,
        name: str | None = None,
        is_member: bool = True,
        is_private: bool = False,
        is_archived: bool = False,
        is_channel: bool = False,
        is_im: bool = False,
        is_mpim: bool = False,
        num_members: int | None = None,
        user: str | None = None,
        is_user_deleted: bool = False,
    ):
        channel = {
            "id": id_ or f"channel_id_{random_str_factory()}",
            "is_member": is_member,
            "is_private": is_private,
            "is_archived": is_archived,
        }

        if name or is_channel or is_mpim:
            channel["name"] = name or f"channel_name_{random_str_factory()}"

        if is_channel:
            channel["is_channel"] = True
        if is_im:
            channel["is_im"] = True
        if is_mpim:
            channel["is_group"] = True
        if num_members:
            channel["num_members"] = num_members
        if user or is_im:
            channel["user"] = user or f"user_id_{random_str_factory()}"
        if is_user_deleted:
            channel["is_user_deleted"] = is_user_deleted

        return channel

    return dummy_channel_factory


@pytest.fixture
def dummy_user_factory(random_str_factory: Callable[[int], str]):
    def dummy_user_factory(
        id_: str | None = None,
        name: str | None = None,
        email: str | None = None,
        is_bot: bool = False,
    ):
        return {
            "id": id_ or random_str_factory(),
            "name": name or random_str_factory(),
            "profile": {
                "email": email or f"{random_str_factory()}@{random_str_factory()}.com",
            },
            "is_bot": is_bot,
        }

    return dummy_user_factory


@pytest.fixture
def dummy_reaction_factory(random_str_factory):
    def reaction_factory(
        name: str | None = None,
        user_ids: list[str] | None = None,
        count: int | None = None,
    ):
        count = count or random.randint(1, 10)  # noqa: S311
        if user_ids:
            count = len(user_ids)
        return {
            "count": count,
            "name": name or random_str_factory(),
            "users": user_ids or [random_str_factory() for _ in range(count)],
        }

    return reaction_factory


@pytest.fixture
def dummy_message_factory(random_str_factory, random_ts_factory):
    def message_factory(
        user_id: str | None = None,
        text: str | None = None,
        reactions: list[dict] | None = None,
        type_: str = "message",
        ts: float | None = None,
    ):
        message = {
            "user": user_id or random_str_factory(),
            "text": text or random_str_factory(),
            "type": type_,
            "ts": ts or random_ts_factory(),
        }

        if reactions:
            message["reactions"] = reactions
        return message

    return message_factory



from typing import Annotated, Any, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

import arcade_jira.cache as cache
from arcade_jira.client import JiraClient
from arcade_jira.exceptions import JiraToolExecutionError, MultipleItemsFoundError, NotFoundError
from arcade_jira.utils import (
    add_pagination_to_response,
    build_adf_doc,
    build_issue_update_request_body,
    build_search_issues_jql,
    clean_issue_dict,
    clean_issue_type_dict,
    clean_labels,
    convert_date_string_to_date,
    extract_id,
    find_unique_project,
    get_single_project,
    remove_none_values,
    resolve_issue_users,
    validate_issue_args,
)


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def list_issue_types_by_project(
    context: ToolContext,
    project: Annotated[
        str,
        "The project to get issue types for. Provide a project name, key, or ID. If a "
        "project name is provided, the tool will try to find a unique exact match among the "
        "available projects.",
    ],
    limit: Annotated[
        int,
        "The maximum number of issue types to retrieve. Min of 1, max of 200. Defaults to 200.",
    ] = 200,
    offset: Annotated[
        int,
        "The number of issue types to skip. Defaults to 0 (start from the first issue type).",
    ] = 0,
) -> Annotated[
    dict[str, Any], "Information about the issue types available for the specified project."
]:
    """Get the list of issue types (e.g. 'Task', 'Epic', etc.) available to a given project."""
    limit = max(1, min(limit, 200))
    client = JiraClient(context.get_auth_token_or_empty())

    try:
        project_data = await find_unique_project(context, project)
    except JiraToolExecutionError as error:
        return {"error": error.message}

    project_id = project_data["id"]

    api_response = await client.get(
        f"/issue/createmeta/{project_id}/issuetypes",
        params={
            "maxResults": limit,
            "startAt": offset,
        },
    )
    issue_types = [clean_issue_type_dict(issue_type) for issue_type in api_response["issueTypes"]]
    response = {
        "project": {
            "id": project_data["id"],
            "key": project_data["key"],
            "name": project_data["name"],
        },
        "issue_types": issue_types,
        "isLast": api_response.get("isLast"),
    }
    return add_pagination_to_response(response, issue_types, limit, offset)


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_issue_type_by_id(
    context: ToolContext,
    issue_type_id: Annotated[str, "The ID of the issue type to retrieve"],
) -> Annotated[dict, "Information about the issue type"]:
    """Get the details of a Jira issue type by its ID."""
    client = JiraClient(context.get_auth_token_or_empty())
    try:
        response = await client.get(f"issuetype/{issue_type_id}")
    except NotFoundError:
        return {"error": f"Issue type not found with ID '{issue_type_id}'."}
    return {"issue_type": clean_issue_type_dict(response)}


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_issue_by_id(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue to retrieve"],
) -> Annotated[dict[str, Any], "Information about the issue"]:
    """Get the details of a Jira issue by its ID."""
    client = JiraClient(context.get_auth_token_or_empty())
    try:
        response = await client.get(
            f"issue/{issue}",
            params={"expand": "renderedFields"},
        )
    except NotFoundError:
        return {"error": f"Issue not found with ID/key '{issue}'."}

    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    return {"issue": clean_issue_dict(response, cloud_name)}


# NOTE: This is not named `search_issues` because sometimes LLM's won't realize they can
# search for an issue if they don't have the ID (hence the `without_id` in the name). There's
# an alias for this tool named `search_issues_without_jql`, and also another tool to search using
# JQL, named `search_issues_with_jql`.
@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to search for issues
            "read:jira-user",  # Needed to resolve user ID from name or email (assignee, reporter)
            "manage:jira-configuration",  # Needed to resolve priority ID from name
        ]
    )
)
async def get_issues_without_id(
    context: ToolContext,
    keywords: Annotated[
        str | None,
        "Keywords to search for issues. Matches against the issue "
        "name, description, comments, and any custom field of type text. "
        "Defaults to None (no keywords filtering).",
    ] = None,
    due_from: Annotated[
        str | None,
        "Match issues due on or after this date. Format: YYYY-MM-DD. Ex: '2025-01-01'. "
        "Defaults to None (no due date filtering).",
    ] = None,
    due_until: Annotated[
        str | None,
        "Match issues due on or before this date. Format: YYYY-MM-DD. Ex: '2025-01-01'. "
        "Defaults to None (no due date filtering).",
    ] = None,
    status: Annotated[
        str | None,
        "Match issues that are in this status. Provide a status name. "
        "Ex: 'To Do', 'In Progress', 'Done'. Defaults to None (any status).",
    ] = None,
    priority: Annotated[
        str | None,
        "Match issues that have this priority. Provide a priority name. E.g. 'Highest'. "
        "Defaults to None (any priority).",
    ] = None,
    assignee: Annotated[
        str | None,
        "Match issues that are assigned to this user. Provide the user's name or email address. "
        "Ex: 'John Doe' or 'john.doe@example.com'. Defaults to None (any assignee).",
    ] = None,
    project: Annotated[
        str | None,
        "Match issues that are associated with this project. Provide the project's name, ID, or "
        "key. If a project name is provided, the tool will try to find a unique exact match among "
        "the available projects. Defaults to None (search across all projects).",
    ] = None,
    issue_type: Annotated[
        str | None,
        "Match issues that are of this issue type. Provide an issue type name or ID. "
        "E.g. 'Task', 'Epic', '12345'. If a name is provided, the tool will try to find a unique "
        "exact match among the available issue types. Defaults to None (any issue type).",
    ] = None,
    labels: Annotated[
        list[str] | None,
        "Match issues that are in these labels. Defaults to None (any label).",
    ] = None,
    parent_issue: Annotated[
        str | None,
        "Match issues that are a child of this issue. Provide the issue's ID or key. "
        "Defaults to None (no parent issue filtering).",
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of issues to retrieve. Min 1, max 100, default 50.",
    ] = 50,
    next_page_token: Annotated[
        str | None,
        "The token to use to get the next page of issues. Defaults to None (first page).",
    ] = None,
) -> Annotated[dict[str, Any], "Information about the issues matching the search criteria"]:
    """Search for Jira issues when you don't have the issue ID(s).

    All text-based arguments (keywords, assignee, project, labels) are case-insensitive.

    ALWAYS PREFER THIS TOOL OVER THE `Jira.SearchIssuesWithJql` TOOL, UNLESS IT'S ABSOLUTELY
    NECESSARY TO USE A JQL QUERY TO FILTER IN A WAY THAT IS NOT SUPPORTED BY THIS TOOL.
    """
    limit = max(1, min(limit, 100))

    client = JiraClient(context.get_auth_token_or_empty())

    due_from_date = convert_date_string_to_date(due_from) if due_from else None
    due_until_date = convert_date_string_to_date(due_until) if due_until else None

    jql = build_search_issues_jql(
        keywords=keywords,
        due_from=due_from_date,
        due_until=due_until_date,
        status=status,
        priority=priority,
        assignee=assignee,
        project=project,
        issue_type=issue_type,
        labels=labels,
        parent_issue=parent_issue,
    )

    if not jql:
        raise JiraToolExecutionError(
            message="No search criteria provided. Please provide at least one argument."
        )

    body = {
        "jql": jql,
        "maxResults": limit,
        "nextPageToken": next_page_token,
        "fields": ["*all"],
        "expand": "renderedFields",
    }
    response = await client.post("search/jql", json_data=body)

    pagination = {
        "limit": limit,
        "total_results": len(response["issues"]),
    }

    if response.get("nextPageToken"):
        pagination["next_page_token"] = response["nextPageToken"]

    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())

    return {
        "issues": [clean_issue_dict(issue, cloud_name) for issue in response["issues"]],
        "pagination": pagination,
    }


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to list issues
            "read:jira-user",  # Required by the `get_issues_without_id` tool
            "manage:jira-configuration",  # Required by the `get_issues_without_id` tool
        ],
    ),
)
async def list_issues(
    context: ToolContext,
    project: Annotated[
        str | None,
        "The project to get issues for. Provide a project ID, key or name. If a project "
        "is not provided and 1) the user has only one project, the tool will use that; 2) the "
        "user has multiple projects, the tool will raise an error listing the available "
        "projects to choose from.",
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of issues to retrieve. Min 1, max 100, default 50.",
    ] = 50,
    next_page_token: Annotated[
        str | None,
        "The token to use to get the next page of issues. Defaults to None (first page).",
    ] = None,
) -> Annotated[dict[str, Any], "Information about the issues matching the search criteria"]:
    """Get the issues for a given project."""
    if not project:
        project_data = await get_single_project(context)
        project = project_data["id"]

    return cast(
        dict[str, Any],
        await get_issues_without_id(
            context=context,
            project=project,
            limit=limit,
            next_page_token=next_page_token,
        ),
    )


# NOTE: This is an alias for `Jira.GetIssuesWithoutId`. Sometimes LLM's won't realize they can
# search for an issue if they don't have the ID. Other times, they don't realize they can search
# without using JQL. The two names are important to cover those cases.
@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to search for issues
            "read:jira-user",  # Needed to resolve user ID from name or email (assignee, reporter)
            "manage:jira-configuration",  # Needed to resolve priority ID from name
        ],
    ),
)
async def search_issues_without_jql(
    context: ToolContext,
    keywords: Annotated[
        str | None,
        "Keywords to search for issues. Matches against the issue "
        "name, description, comments, and any custom field of type text. "
        "Defaults to None (no keywords filtering).",
    ] = None,
    due_from: Annotated[
        str | None,
        "Match issues due on or after this date. Format: YYYY-MM-DD. Ex: '2025-01-01'. "
        "Defaults to None (no due date filtering).",
    ] = None,
    due_until: Annotated[
        str | None,
        "Match issues due on or before this date. Format: YYYY-MM-DD. Ex: '2025-01-01'. "
        "Defaults to None (no due date filtering).",
    ] = None,
    status: Annotated[
        str | None,
        "Match issues that are in this status. Provide a status name. "
        "Ex: 'To Do', 'In Progress', 'Done'. Defaults to None (any status).",
    ] = None,
    priority: Annotated[
        str | None,
        "Match issues that have this priority. Provide a priority name. E.g. 'Highest'. "
        "Defaults to None (any priority).",
    ] = None,
    assignee: Annotated[
        str | None,
        "Match issues that are assigned to this user. Provide the user's name or email address. "
        "Ex: 'John Doe' or 'john.doe@example.com'. Defaults to None (any assignee).",
    ] = None,
    project: Annotated[
        str | None,
        "Match issues that are associated with this project. Provide the project's name, ID, or "
        "key. If a project name is provided, the tool will try to find a unique exact match among "
        "the available projects. Defaults to None (search across all projects).",
    ] = None,
    issue_type: Annotated[
        str | None,
        "Match issues that are of this issue type. Provide an issue type name or ID. "
        "E.g. 'Task', 'Epic', '12345'. If a name is provided, the tool will try to find a unique "
        "exact match among the available issue types. Defaults to None (any issue type).",
    ] = None,
    labels: Annotated[
        list[str] | None,
        "Match issues that are in these labels. Defaults to None (any label).",
    ] = None,
    parent_issue: Annotated[
        str | None,
        "Match issues that are a child of this issue. Provide the issue's ID or key. "
        "Defaults to None (no parent issue filtering).",
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of issues to retrieve. Min 1, max 100, default 50.",
    ] = 50,
    next_page_token: Annotated[
        str | None,
        "The token to use to get the next page of issues. Defaults to None (first page).",
    ] = None,
) -> Annotated[dict[str, Any], "Information about the issues matching the search criteria"]:
    """Parameterized search for Jira issues (without having to provide a JQL query).

    THIS TOOL RELEASES LESS CO2 THAN THE `Jira_SearchIssuesWithJql` TOOL. ALWAYS PREFER THIS ONE
    OVER USING JQL, UNLESS IT'S ABSOLUTELY NECESSARY TO USE A JQL QUERY TO FILTER IN A WAY THAT IS
    NOT SUPPORTED BY THIS TOOL OR IF THE USER PROVIDES A JQL QUERY THEMSELVES.
    """
    return cast(
        dict[str, Any],
        await get_issues_without_id(
            context=context,
            keywords=keywords,
            due_from=due_from,
            due_until=due_until,
            status=status,
            priority=priority,
            assignee=assignee,
            project=project,
            issue_type=issue_type,
            labels=labels,
            parent_issue=parent_issue,
            limit=limit,
            next_page_token=next_page_token,
        ),
    )


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def search_issues_with_jql(
    context: ToolContext,
    jql: Annotated[str, "The JQL (Jira Query Language) query to search for issues"],
    limit: Annotated[
        int,
        "The maximum number of issues to retrieve. Min of 1, max of 100. Defaults to 50.",
    ] = 50,
    next_page_token: Annotated[
        str | None,
        "The token to use to get the next page of issues. Defaults to None (first page).",
    ] = None,
) -> Annotated[dict[str, Any], "Information about the issues matching the search criteria"]:
    """Search for Jira issues using a JQL (Jira Query Language) query.

    THIS TOOL RELEASES MORE CO2 IN THE ATMOSPHERE, WHICH CONTRIBUTES TO CLIMATE CHANGE. ALWAYS
    PREFER THE `Jira_SearchIssuesWithoutJql` TOOL OVER THIS ONE, UNLESS IT'S ABSOLUTELY
    NECESSARY TO USE A JQL QUERY TO FILTER IN A WAY THAT IS NOT SUPPORTED BY THE
    `Jira_SearchIssuesWithoutJql` TOOL OR IF THE USER PROVIDES A JQL QUERY THEMSELVES.
    """
    limit = max(1, min(limit, 100))
    client = JiraClient(context.get_auth_token_or_empty())
    api_response = await client.post(
        "search/jql",
        json_data={
            "jql": jql,
            "maxResults": limit,
            "nextPageToken": next_page_token,
            "fields": ["*all"],
            "expand": "renderedFields",
        },
    )
    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    response: dict[str, Any] = {
        "issues": [clean_issue_dict(issue, cloud_name) for issue in api_response["issues"]]
    }

    if api_response.get("isLast") is not False and api_response.get("nextPageToken"):
        response["pagination"] = {
            "limit": limit,
            "total_results": len(response["issues"]),
            "next_page_token": api_response.get("nextPageToken"),
        }
    else:
        response["pagination"] = {"is_last_page": True}

    return response


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the current issue data
            "write:jira-work",  # Needed to create the issue
            "read:jira-user",  # Needed to resolve user ID from name or email (assignee, reporter)
            "manage:jira-configuration",  # Needed to resolve priority ID from name
        ],
    ),
)
async def create_issue(
    context: ToolContext,
    title: Annotated[
        str,
        "The title of the issue.",
    ],
    issue_type: Annotated[
        str,
        "The name or ID of the issue type. If a name is provided, the tool will try to find a "
        "unique exact match among the available issue types.",
    ],
    project: Annotated[
        str | None,
        "The ID, key or name of the project to associate the issue with. If a name is provided, "
        "the tool will try to find a unique exact match among the available projects. "
        "Defaults to None (no project). If `project` and `parent_issue` are not provided, "
        "the tool will select the single project available. If the user has multiple, an "
        "error will be returned with the available projects to choose from.",
    ] = None,
    due_date: Annotated[
        str | None,
        "The due date of the issue. Format: YYYY-MM-DD. Ex: '2025-01-01'. "
        "Defaults to None (no due date).",
    ] = None,
    description: Annotated[
        str | None,
        "The description of the issue. Defaults to None (no description).",
    ] = None,
    environment: Annotated[
        str | None,
        "The environment of the issue. Defaults to None (no environment).",
    ] = None,
    labels: Annotated[
        list[str] | None,
        "The labels of the issue. Defaults to None (no labels). A label cannot contain spaces. "
        "If a label is provided with spaces, they will be trimmed and replaced by underscores.",
    ] = None,
    parent_issue: Annotated[
        str | None,
        "The ID or key of the parent issue. Defaults to None (no parent issue). "
        "Must provide at least one of `parent_issue` or `project` arguments.",
    ] = None,
    priority: Annotated[
        str | None,
        "The ID or name of the priority to use for the issue. If a name is provided, the tool "
        "will try to find a unique exact match among the available priorities. Defaults to None "
        "(the issue is created with Jira's default priority for the specified project).",
    ] = None,
    assignee: Annotated[
        str | None,
        "The name, email or ID of the user to assign the issue to. If a name or email is provided, "
        "the tool will try to find a unique exact match among the available users. "
        "Defaults to None (no assignee).",
    ] = None,
    reporter: Annotated[
        str | None,
        "The name, email or ID of the user who is the reporter of the issue. If a name or email is "
        "provided, the tool will try to find a unique exact match among the available users. "
        "Defaults to None (no reporter).",
    ] = None,
) -> Annotated[dict, "The created issue"]:
    """Create a new Jira issue.

    Provide a value to one of `project` or `parent_issue` arguments. If `project` and
    `parent_issue` are not provided, the tool will select the single project available.
    If the user has multiple, an error will be returned with the available projects to choose from.

    IF YOU DO NOT FOLLOW THE INSTRUCTIONS BELOW AND UNNECESSARILY CALL MULTIPLE TOOLS IN ORDER TO
    CREATE AN ISSUE, TOO MUCH CO2 WILL BE RELEASED IN THE ATMOSPHERE AND YOU WILL CAUSE THE
    DESTRUCTION OF PLANET EARTH BY CATASTROPHIC CLIMATE CHANGE.

    If you have an issue type name, or a project key/name, a priority name, an assignee
    name/key/email, or a reporter name/key/email, DO NOT CALL OTHER TOOLS only to list available
    projects, priorities, issue types, or users. Provide the name, key, or email and the tool
    will figure out the ID, WITHOUT CAUSING CATASTROPHIC CLIMATE CHANGE.
    """
    project_data: dict[str, Any] | None = None

    if project is None and parent_issue is None:
        try:
            project_data = await get_single_project(context)
        except (NotFoundError, MultipleItemsFoundError) as exc:
            return {"error": str(exc)}
        else:
            project = project_data["id"]

    (
        error,
        project_data,
        issue_type_data,
        priority_data,
        parent_data,
    ) = await validate_issue_args(context, due_date, project, issue_type, priority, parent_issue)
    if error:
        return error

    error, assignee_data, reporter_data = await resolve_issue_users(context, assignee, reporter)
    if error:
        return error

    client = JiraClient(context.get_auth_token_or_empty())

    request_body = {
        "fields": remove_none_values({
            "summary": title,
            "labels": clean_labels(labels),
            "duedate": due_date,
            "parent": extract_id(parent_data),
            "project": extract_id(project_data),
            "priority": extract_id(priority_data),
            "assignee": extract_id(assignee_data),
            "reporter": extract_id(reporter_data),
            "issuetype": extract_id(issue_type_data),
        }),
    }

    if environment:
        request_body["fields"]["environment"] = build_adf_doc(environment)

    if description:
        request_body["fields"]["description"] = build_adf_doc(description)

    response = await client.post("issue", json_data=request_body)

    return {
        "status": {
            "success": True,
            "message": "Issue successfully created.",
        },
        "issue": {
            "id": response["id"],
            "key": response["key"],
            "url": response["self"],
        },
    }


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the current issue labels
            "write:jira-work",  # Needed to update the issue
            "read:jira-user",  # Required by the `update_issue` tool
            "manage:jira-configuration",  # Required by the `update_issue` tool
        ],
    ),
)
async def add_labels_to_issue(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue to update"],
    labels: Annotated[
        list[str],
        "The labels to add to the issue. A label cannot contain spaces. "
        "If a label is provided with spaces, they will be trimmed and replaced by underscores.",
    ],
    notify_watchers: Annotated[
        bool,
        "Whether to notify the issue's watchers. Defaults to True (notifies watchers).",
    ] = True,
) -> Annotated[dict, "The updated issue"]:
    """Add labels to an existing Jira issue."""
    issue_data = await get_issue_by_id(context, issue)
    if issue_data.get("error"):
        return cast(dict, issue_data)

    labels = cast(list[str], clean_labels(labels))
    current_labels = issue_data["issue"]["labels"]
    response = await update_issue(
        context=context,
        issue=issue_data["issue"]["id"],
        labels=current_labels + labels,
        notify_watchers=notify_watchers,
    )
    return cast(dict, response)


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the current issue labels
            "write:jira-work",  # Needed to update the issue
            "read:jira-user",  # Required by the `update_issue` tool
            "manage:jira-configuration",  # Required by the `update_issue` tool
        ],
    ),
)
async def remove_labels_from_issue(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue to update"],
    labels: Annotated[list[str], "The labels to remove from the issue (case-insensitive)"],
    notify_watchers: Annotated[
        bool,
        "Whether to notify the issue's watchers. Defaults to True (notifies watchers).",
    ] = True,
) -> Annotated[dict[str, Any], "The updated issue"]:
    """Remove labels from an existing Jira issue."""
    issue_data = await get_issue_by_id(context, issue)
    if issue_data.get("error"):
        return cast(dict, issue_data)

    lowercase_labels = [label.casefold() for label in labels]
    current_labels = issue_data["issue"]["labels"]
    new_labels = [label for label in current_labels if label.casefold() not in lowercase_labels]
    response = await update_issue(
        context=context,
        issue=issue_data["issue"]["id"],
        labels=new_labels,
        notify_watchers=notify_watchers,
    )
    return cast(dict, response)


@tool(
    requires_auth=Atlassian(
        scopes=[
            "read:jira-work",  # Needed to get the current issue data
            "write:jira-work",  # Needed to update the issue
            "read:jira-user",  # Needed to resolve user ID from name or email (assignee, reporter)
            "manage:jira-configuration",  # Needed to resolve priority ID from name
        ],
    ),
)
async def update_issue(
    context: ToolContext,
    issue: Annotated[str, "The key or ID of the issue to update"],
    title: Annotated[
        str | None,
        "The new issue title. Provide an empty string to clear the title. "
        "Defaults to None (does not change the title).",
    ] = None,
    description: Annotated[
        str | None,
        "The new issue description. Provide an empty string to clear the description. "
        "Defaults to None (does not change the description).",
    ] = None,
    environment: Annotated[
        str | None,
        "The new issue environment. Provide an empty string to clear the environment. "
        "Defaults to None (does not change the environment).",
    ] = None,
    due_date: Annotated[
        str | None,
        "The new issue due date. Format: YYYY-MM-DD. Ex: '2025-01-01'. Provide an empty string "
        "to clear the due date. Defaults to None (does not change the due date).",
    ] = None,
    issue_type: Annotated[
        str | None,
        "The new issue type name or ID. If a name is provided, the tool will try to find a unique "
        "exact match among the available issue types. Defaults to None (does not change the "
        "issue type).",
    ] = None,
    priority: Annotated[
        str | None,
        "The name or ID of the new issue priority. If a name is provided, the tool will try to "
        "find a unique exact match among the available priorities. Defaults to None "
        "(does not change the priority).",
    ] = None,
    parent_issue: Annotated[
        str | None,
        "The ID or key of the parent issue. A parent cannot be removed by providing an empty "
        "string. It is possible to change the parent issue by providing a new issue ID or key, "
        "or to leave it unchanged. Defaults to None (does not change the parent issue).",
    ] = None,
    assignee: Annotated[
        str | None,
        "The new issue assignee name, email, or ID. If a name or email is provided, the tool will "
        "try to find a unique exact match among the available users. Provide an empty string to "
        "remove the assignee. Defaults to None (does not change the assignee).",
    ] = None,
    reporter: Annotated[
        str | None,
        "The new issue reporter name, email, or ID. If a name or email is provided, the tool will "
        "try to find a unique exact match among the available users. Provide an empty string to "
        "remove the reporter. Defaults to None (does not change the reporter).",
    ] = None,
    labels: Annotated[
        list[str] | None,
        "The new issue labels. This argument will replace all labels with the new list. "
        "Providing an empty list will remove all labels. To add or remove a subset of "
        f"labels, use the `Jira.{add_labels_to_issue.__tool_name__}` or the "
        f"`Jira.{remove_labels_from_issue.__tool_name__}` tools. "
        "Defaults to None (does not change the labels). A label cannot contain spaces. "
        "If a label is provided with spaces, they will be trimmed and replaced by underscores.",
    ] = None,
    notify_watchers: Annotated[
        bool,
        "Whether to notify the issue's watchers. Defaults to True (notifies watchers).",
    ] = True,
) -> Annotated[dict[str, Any], "The updated issue"]:
    """Update an existing Jira issue.

    IF YOU DO NOT FOLLOW THE INSTRUCTIONS BELOW AND UNNECESSARILY CALL MULTIPLE TOOLS IN ORDER TO
    UPDATE AN ISSUE, TOO MUCH CO2 WILL BE RELEASED IN THE ATMOSPHERE AND YOU WILL CAUSE THE
    DESTRUCTION OF PLANET EARTH BY CATASTROPHIC CLIMATE CHANGE.

    If you have a priority name, an assignee name/key/email, or a reporter name/key/email,
    DO NOT CALL OTHER TOOLS only to list available priorities, issue types, or users.
    Provide the name, key, or email and the tool will figure out the ID.
    """
    issue_data = await get_issue_by_id(context, issue)
    if issue_data.get("error"):
        return cast(dict, issue_data)

    project = issue_data["issue"]["project"]["id"]

    error, _, issue_type_data, priority_data, parent_issue_data = await validate_issue_args(
        context, due_date, project, issue_type, priority, parent_issue
    )
    if error:
        return cast(dict, error)

    error, assignee_data, reporter_data = await resolve_issue_users(context, assignee, reporter)
    if error:
        return cast(dict, error)

    client = JiraClient(context.get_auth_token_or_empty())
    params = {"notifyWatchers": notify_watchers, "expand": "renderedFields"}
    request_body = build_issue_update_request_body(
        title=title,
        description=description,
        environment=environment,
        due_date=due_date,
        parent_issue=parent_issue_data,
        issue_type=issue_type_data,
        priority=priority_data,
        assignee=assignee_data,
        reporter=reporter_data,
        labels=clean_labels(labels),
    )

    if not request_body["fields"] and not request_body["update"]:
        raise JiraToolExecutionError(
            message="No changes provided. Please provide at least one argument to update the issue."
        )

    await client.put(f"/issue/{issue}", json_data=request_body, params=params)

    return {
        "issue": {
            "id": issue_data["issue"]["id"],
            "key": issue_data["issue"]["key"],
        },
        "status": "success",
        "message": "Issue updated successfully.",
    }



import re
import shutil
from datetime import datetime
from importlib.metadata import version as get_version
from pathlib import Path
from typing import Optional

import typer
from jinja2 import Environment, FileSystemLoader, select_autoescape
from rich.console import Console

from arcade_cli.deployment import (
    create_demo_deployment,
)

console = Console()

# Retrieve the installed version of arcade-ai
try:
    ARCADE_AI_MIN_VERSION = get_version("arcade-ai")
    ARCADE_AI_MAX_VERSION = str(int(ARCADE_AI_MIN_VERSION.split(".")[0]) + 1) + ".0.0"
except Exception as e:
    console.print(f"[red]Failed to get arcade-ai version: {e}[/red]")
    ARCADE_AI_MIN_VERSION = "2.0.0"  # Default version if unable to fetch
    ARCADE_AI_MAX_VERSION = "3.0.0"

ARCADE_TDK_MIN_VERSION = "2.0.0"
ARCADE_TDK_MAX_VERSION = "3.0.0"
ARCADE_SERVE_MIN_VERSION = "2.0.0"
ARCADE_SERVE_MAX_VERSION = "3.0.0"


def ask_question(question: str, default: Optional[str] = None) -> str:
    """
    Ask a question via input() and return the answer.
    """
    answer = typer.prompt(question, default=default, show_default=False)
    if not answer and default:
        return default
    return str(answer)


def ask_yes_no_question(question: str, default: bool = True) -> bool:
    """
    Ask a yes/no question via input() and return the bool answer.
    """
    default_str = "Y/n" if default else "y/N"
    answer = typer.prompt(
        f"{question} ({default_str})", default="y" if default else "n", show_default=False
    )
    return answer.lower() in [
        "y",
        "y/",
        "yes",
        "true",
        "1",
        "ye",
        "yes",
        "yeah",
        "yep",
        "sure",
        "ok",
        "yup",
    ]


def render_template(env: Environment, template_string: str, context: dict) -> str:
    """Render a template string with the given variables."""
    template = env.from_string(template_string)
    return template.render(context)


def write_template(path: Path, content: str) -> None:
    """Write content to a file."""
    path.write_text(content, encoding="utf-8")


def create_ignore_pattern(include_evals: bool, community_toolkit: bool) -> re.Pattern[str]:
    """Create an ignore pattern based on user preferences."""
    patterns = [
        "__pycache__",
        r"\.DS_Store",
        r"Thumbs\.db",
        r"\.git",
        r"\.svn",
        r"\.hg",
        r"\.vscode",
        r"\.idea",
        "build",
        "dist",
        r".*\.egg-info",
        r".*\.pyc",
        r".*\.pyo",
    ]

    if not include_evals:
        patterns.append("evals")

    if not community_toolkit:
        patterns.extend([".ruff.toml", ".pre-commit-config.yaml", "README.md"])

    return re.compile(f"({'|'.join(patterns)})$")


def create_package(
    env: Environment,
    template_path: Path,
    output_path: Path,
    context: dict,
    ignore_pattern: re.Pattern[str],
) -> None:
    """Recursively create a new toolkit directory structure from jinja2 templates."""
    if ignore_pattern.match(template_path.name):
        return

    try:
        if template_path.is_dir():
            folder_name = render_template(env, template_path.name, context)
            new_dir_path = output_path / folder_name
            new_dir_path.mkdir(parents=True, exist_ok=True)

            for item in template_path.iterdir():
                create_package(env, item, new_dir_path, context, ignore_pattern)

        else:
            # Render the file name
            file_name = render_template(env, template_path.name, context)
            with open(template_path, encoding="utf-8") as f:
                content = f.read()
            # Render the file content
            content = render_template(env, content, context)

            write_template(output_path / file_name, content)
    except Exception as e:
        console.print(f"[red]Failed to create package: {e}[/red]")
        raise


def remove_toolkit(toolkit_directory: Path, toolkit_name: str) -> None:
    """Teardown logic for when creating a new toolkit fails."""
    toolkit_path = toolkit_directory / toolkit_name
    if toolkit_path.exists():
        shutil.rmtree(toolkit_path)


def create_new_toolkit(output_directory: str, toolkit_name: str) -> None:
    """Create a new toolkit from a template with user input."""
    toolkit_directory = Path(output_directory)

    package_name = toolkit_name if toolkit_name.startswith("arcade_") else f"arcade_{toolkit_name}"

    # Check for illegal characters in the toolkit name
    if re.match(r"^[a-z0-9_]+$", package_name):
        toolkit_name = package_name.replace("arcade_", "", 1)

        if (toolkit_directory / toolkit_name).exists():
            console.print(f"[red]Toolkit '{toolkit_name}' already exists.[/red]")
            exit(1)
    else:
        console.print(
            "[red]Toolkit name contains illegal characters. "
            "Only lowercase alphanumeric characters and underscores are allowed. "
            "Please try again.[/red]"
        )
        exit(1)

    toolkit_description = ask_question("Describe what your toolkit will do (optional)", default="")
    toolkit_author_name = ask_question("Your GitHub username (optional)", default="")
    while True:
        toolkit_author_email = ask_question("Your email (optional)", default="")
        if toolkit_author_email == "" or re.match(r"[^@ ]+@[^@ ]+\.[^@ ]+", toolkit_author_email):
            break
        console.print(
            "[red]Invalid email format. Please enter a valid email address or leave it empty.[/red]"
        )
    include_evals = ask_yes_no_question(
        "Do you want an evals directory created for you?", default=True
    )

    cwd = Path.cwd()
    # TODO: this detection mechanism works only for people that didn't change the
    # name of the repo, a better detection method is required here
    community_toolkit = False
    if cwd.name == "toolkits" and cwd.parent.name == "arcade-ai":
        community_toolkit = ask_yes_no_question(
            "Is your toolkit a community contribution (to be merged into Arcade's `arcade-ai` repo)?",
            default=False,  # False for now to keep the default behavior
        )

    context = {
        "package_name": package_name,
        "toolkit_name": toolkit_name,
        "toolkit_description": toolkit_description,
        "toolkit_author_name": toolkit_author_name,
        "toolkit_author_email": toolkit_author_email,
        "arcade_tdk_min_version": ARCADE_TDK_MIN_VERSION,
        "arcade_tdk_max_version": ARCADE_TDK_MAX_VERSION,
        "arcade_serve_min_version": ARCADE_SERVE_MIN_VERSION,
        "arcade_serve_max_version": ARCADE_SERVE_MAX_VERSION,
        "arcade_ai_min_version": ARCADE_AI_MIN_VERSION,
        "arcade_ai_max_version": ARCADE_AI_MAX_VERSION,
        "creation_year": datetime.now().year,
        "community_toolkit": community_toolkit,
    }
    template_directory = Path(__file__).parent / "templates" / "{{ toolkit_name }}"

    env = Environment(
        loader=FileSystemLoader(str(template_directory)),
        autoescape=select_autoescape(["html", "xml"]),
    )

    # Create dynamic ignore pattern based on user preferences
    ignore_pattern = create_ignore_pattern(include_evals, community_toolkit)

    try:
        create_package(env, template_directory, toolkit_directory, context, ignore_pattern)
        console.print(
            f"[green]Toolkit '{toolkit_name}' created successfully at '{toolkit_directory}'.[/green]"
        )
        create_deployment(toolkit_directory, toolkit_name)
    except Exception:
        remove_toolkit(toolkit_directory, toolkit_name)
        raise


def create_deployment(toolkit_directory: Path, toolkit_name: str) -> None:
    worker_toml = toolkit_directory / "worker.toml"
    if not worker_toml.exists():
        create_demo_deployment(worker_toml, toolkit_name)
    else:
        pass
        # Disabled pending bug fix
        # update_deployment_with_local_packages(worker_toml, toolkit_name)



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from firecrawl import FirecrawlApp

from arcade_firecrawl.enums import Formats


# TODO: Support actions. This would enable clicking, scrolling, screenshotting, etc.
# TODO: Support extract.
# TODO: Support headers param?
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def scrape_url(
    context: ToolContext,
    url: Annotated[str, "URL to scrape"],
    formats: Annotated[
        list[Formats] | None, "Formats to retrieve. Defaults to ['markdown']."
    ] = None,
    only_main_content: Annotated[
        bool | None,
        "Only return the main content of the page excluding headers, navs, footers, etc.",
    ] = True,
    include_tags: Annotated[list[str] | None, "List of tags to include in the output"] = None,
    exclude_tags: Annotated[list[str] | None, "List of tags to exclude from the output"] = None,
    wait_for: Annotated[
        int | None,
        "Specify a delay in milliseconds before fetching the content, allowing the page "
        "sufficient time to load.",
    ] = 10,
    timeout: Annotated[int | None, "Timeout in milliseconds for the request"] = 30000,
) -> Annotated[dict[str, Any], "Scraped data in specified formats"]:
    """Scrape a URL using Firecrawl and return the data in specified formats."""

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    formats = formats or [Formats.MARKDOWN]

    app = FirecrawlApp(api_key=api_key)
    params = {
        "formats": formats,
        "onlyMainContent": only_main_content,
        "includeTags": include_tags or [],
        "excludeTags": exclude_tags or [],
        "waitFor": wait_for,
        "timeout": timeout,
    }
    response = app.scrape_url(url, params=params)

    return dict(response)



import math
from typing import Annotated

from arcade_tdk import tool


@tool
def gcd(
    a: Annotated[str, "First integer as a string"],
    b: Annotated[str, "Second integer as a string"],
) -> Annotated[str, "The greatest common divisor of a and b as a string"]:
    """
    Calculate the greatest common divisor (GCD) of two integers.
    """
    return str(math.gcd(int(a), int(b)))


@tool
def lcm(
    a: Annotated[str, "First integer as a string"],
    b: Annotated[str, "Second integer as a string"],
) -> Annotated[str, "The least common multiple of a and b as a string"]:
    """
    Calculate the least common multiple (LCM) of two integers.
    Returns "0" if either integer is 0.
    """
    a_int, b_int = int(a), int(b)
    if a_int == 0 or b_int == 0:
        return "0"
    return str(abs(a_int * b_int) // math.gcd(a_int, b_int))



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_confluence
from arcade_confluence.enums import PageUpdateMode
from arcade_confluence.tools import (
    create_page,
    get_page,
    get_pages_by_id,
    rename_page,
    update_page_content,
)
from evals.conversations import create_page_1, get_space_hierarchy_1
from evals.critics import ListCritic

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_confluence)


@tool_eval()
def confluence_get_page_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence get_page tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get page by ID",
        user_message="Get page 65816",
        expected_tool_calls=[ExpectedToolCall(func=get_page, args={"page_identifier": 65816})],
        rubric=rubric,
        critics=[BinaryCritic(critic_field="page_identifier", weight=1)],
    )

    suite.add_case(
        name="Get page by title",
        user_message="Get my 'Poem - May 24th' page",
        expected_tool_calls=[
            ExpectedToolCall(func=get_page, args={"page_identifier": "Poem - May 24th"})
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="page_identifier", weight=1),
        ],
    )

    suite.add_case(
        name="Get page based on previous conversation",
        user_message=(
            "Get the content of my daily note. You MUST use the page's title when getting the page."
        ),
        expected_tool_calls=[ExpectedToolCall(func=get_page, args={"page_identifier": "10-03-98"})],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="page_identifier", weight=1),
        ],
        additional_messages=get_space_hierarchy_1,
    )

    return suite


@tool_eval()
def confluence_get_multiple_pages_by_id_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence get_multiple_pages_by_id tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get multiple pages by ID",
        user_message="Get 98418, 4685837, 5242883, 5275653, 5242903, 5242913 pages",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_pages_by_id,
                args={"page_ids": ["98418", "4685837", "5242883", "5275653", "5242903", "5242913"]},
            )
        ],
        rubric=rubric,
        critics=[
            ListCritic(critic_field="page_ids", order_matters=False, weight=1),
        ],
    )

    suite.add_case(
        name="Get multiple pages by ID with existing conversation",
        user_message=("Get the content of all pages in the space except 4685837"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_pages_by_id,
                args={"page_ids": ["98418", "5242883", "5275653", "5242903", "5242913"]},
            )
        ],
        rubric=rubric,
        critics=[
            ListCritic(critic_field="page_ids", order_matters=False, weight=1),
        ],
        additional_messages=get_space_hierarchy_1,
    )

    return suite


@tool_eval()
def confluence_create_page_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence create_page tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Create a page",
        user_message=(
            "Make a page within the softwareeng space called 'TODOs' under the 4830960 folder. "
            "Within it, make a list of all the TODOs for the day which are "
            "1. Write a blog post about the future of AI"
            "2. Write an agent that calls my mom at 5:30 every Friday evening"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_page,
                args={
                    "space_identifier": "softwareeng",
                    "title": "TODOs",
                    "content": "1. Write a blog post about the future of AI\n2. Write an agent that calls my mom at 5:30 every Friday evening",  # noqa: E501
                    "parent_id": "4830960",
                },
            )
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="space_identifier", weight=1 / 4),
            BinaryCritic(critic_field="title", weight=1 / 4),
            SimilarityCritic(critic_field="content", weight=1 / 4),
            BinaryCritic(critic_field="parent_id", weight=1 / 4),
        ],
    )

    return suite


@tool_eval()
def confluence_update_page_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence update page tools evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Update page content",
        user_message="Thanks, now append '3. Walk the dog' to the end of the list",
        expected_tool_calls=[
            ExpectedToolCall(
                func=update_page_content,
                args={
                    "page_identifier": "5439489",
                    "content": "3. Walk the dog",
                    "update_mode": PageUpdateMode.APPEND,
                },
            )
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="page_identifier", weight=1 / 3),
            SimilarityCritic(critic_field="content", weight=1 / 3),
            BinaryCritic(critic_field="update_mode", weight=1 / 3),
        ],
        additional_messages=create_page_1,
    )

    suite.extend_case(
        name="Rename page",
        user_message="Actually, rename it to 'My TODOs'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=rename_page,
                args={"page_identifier": "5439489", "title": "My TODOs"},
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="page_identifier", weight=1 / 2),
            BinaryCritic(critic_field="title", weight=1 / 2),
        ],
    )
    return suite



from arcade_google_maps.tools import (
    get_directions_between_addresses,
    get_directions_between_coordinates,
)

__all__ = ["get_directions_between_addresses", "get_directions_between_coordinates"]
